# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Dictionary containing commutation relations of unparameterizable standard gates. The dictionary key
is defined by a pair of gates and, optionally, their relative placement to each other, i.e.:

   ┌───┐┌───┐
q: ┤ X ├┤ Z ├    -> standard_gates_commutations["x", "z"] = False
   └───┘└───┘

or
          ┌───┐
q_0: ──■──┤ X ├
     ┌─┴─┐└─┬─┘  -> standard_gates_commutations["cx", "cy"][1, 0] = False
q_1: ┤ Y ├──■──
     └───┘

or

q_0: ──■───────
     ┌─┴─┐
q_1: ┤ X ├──■──  -> standard_gates_commutations['cx', 'cx'][None, 0] = False
     └───┘┌─┴─┐
q_2: ─────┤ X ├
          └───┘
Note that the commutation of a pair of gates is only stored in one specific order, i.e. the commutation
of a pair of cy and cx gates is only stored in standard_gates_commutations["cx", "cy"]. The order of the
gates is derived from the number of qubits in each gate and the value of the integer representation of
the gate name.

The values of standard_gates_commutations can be a Boolean value denoting whether the pair of gates
commute or another dictionary if the relative placement of the pair of gates has an impact on the
commutation. The relative placement is defined by the gate qubit arrangements as q2^{-1}[q1[i]] where
q1[i] is the ith qubit of the first gate and q2^{-1}[q] returns the qubit index of qubit q in the second
gate (possibly 'None'). In the second example, the zeroth qubit of cy overlaps with qubit 1 of cx. Non-
overlapping qubits are specified as 'None' in the dictionary key, e.g. example 3, there is no overlap
on the zeroth qubit of the first cx but the zeroth qubit of the second gate overlaps with qubit 1 of the
first cx.
"""


standard_gates_commutations = {
    ("id", "id"): True,
    ("id", "sx"): True,
    ("id", "cx"): True,
    ("id", "c3sx"): True,
    ("id", "ccx"): True,
    ("id", "dcx"): True,
    ("id", "ch"): True,
    ("id", "cswap"): True,
    ("id", "csx"): True,
    ("id", "cy"): True,
    ("id", "cz"): True,
    ("id", "ccz"): True,
    ("id", "rccx"): True,
    ("id", "rcccx"): True,
    ("id", "ecr"): True,
    ("id", "sdg"): True,
    ("id", "cs"): True,
    ("id", "csdg"): True,
    ("id", "swap"): True,
    ("id", "iswap"): True,
    ("id", "sxdg"): True,
    ("id", "tdg"): True,
    ("id", "rxx"): True,
    ("id", "ryy"): True,
    ("id", "rzz"): True,
    ("id", "rzx"): True,
    ("sx", "sx"): True,
    ("sx", "cx"): {
        (0,): False,
        (1,): True,
    },
    ("sx", "c3sx"): {
        (0,): False,
        (1,): False,
        (2,): False,
        (3,): True,
    },
    ("sx", "ccx"): {
        (0,): False,
        (1,): False,
        (2,): True,
    },
    ("sx", "dcx"): False,
    ("sx", "ch"): False,
    ("sx", "cswap"): False,
    ("sx", "csx"): {
        (0,): False,
        (1,): True,
    },
    ("sx", "cy"): False,
    ("sx", "cz"): False,
    ("sx", "ccz"): False,
    ("sx", "rccx"): False,
    ("sx", "rcccx"): False,
    ("sx", "ecr"): {
        (0,): False,
        (1,): True,
    },
    ("sx", "sdg"): False,
    ("sx", "cs"): False,
    ("sx", "csdg"): False,
    ("sx", "swap"): False,
    ("sx", "iswap"): False,
    ("sx", "sxdg"): True,
    ("sx", "tdg"): False,
    ("sx", "rxx"): True,
    ("sx", "ryy"): False,
    ("sx", "rzz"): False,
    ("sx", "rzx"): {
        (0,): False,
        (1,): True,
    },
    ("x", "id"): True,
    ("x", "sx"): True,
    ("x", "x"): True,
    ("x", "cx"): {
        (0,): False,
        (1,): True,
    },
    ("x", "c3sx"): {
        (0,): False,
        (1,): False,
        (2,): False,
        (3,): True,
    },
    ("x", "ccx"): {
        (0,): False,
        (1,): False,
        (2,): True,
    },
    ("x", "dcx"): False,
    ("x", "ch"): False,
    ("x", "cswap"): False,
    ("x", "csx"): {
        (0,): False,
        (1,): True,
    },
    ("x", "cy"): False,
    ("x", "cz"): False,
    ("x", "ccz"): False,
    ("x", "rccx"): False,
    ("x", "rcccx"): False,
    ("x", "ecr"): {
        (0,): False,
        (1,): True,
    },
    ("x", "sdg"): False,
    ("x", "cs"): False,
    ("x", "csdg"): False,
    ("x", "swap"): False,
    ("x", "iswap"): False,
    ("x", "sxdg"): True,
    ("x", "tdg"): False,
    ("x", "y"): False,
    ("x", "z"): False,
    ("x", "rxx"): True,
    ("x", "ryy"): False,
    ("x", "rzz"): False,
    ("x", "rzx"): {
        (0,): False,
        (1,): True,
    },
    ("cx", "cx"): {
        (0, 1): True,
        (0, None): True,
        (1, 0): False,
        (1, None): False,
        (None, 0): False,
        (None, 1): True,
    },
    ("cx", "c3sx"): {
        (0, 1): False,
        (0, 2): False,
        (0, 3): True,
        (0, None): True,
        (1, 0): False,
        (1, 2): False,
        (1, 3): True,
        (1, None): True,
        (2, 0): False,
        (2, 1): False,
        (2, 3): True,
        (2, None): True,
        (3, 0): False,
        (3, 1): False,
        (3, 2): False,
        (3, None): False,
        (None, 0): False,
        (None, 1): False,
        (None, 2): False,
        (None, 3): True,
    },
    ("cx", "ccx"): {
        (0, 1): False,
        (0, 2): True,
        (0, None): True,
        (1, 0): False,
        (1, 2): True,
        (1, None): True,
        (2, 0): False,
        (2, 1): False,
        (2, None): False,
        (None, 0): False,
        (None, 1): False,
        (None, 2): True,
    },
    ("cx", "dcx"): False,
    ("cx", "cswap"): {
        (0, 1): False,
        (0, 2): False,
        (0, None): True,
        (1, 0): False,
        (1, 2): False,
        (1, None): False,
        (2, 0): False,
        (2, 1): False,
        (2, None): False,
        (None, 0): False,
        (None, 1): False,
        (None, 2): False,
    },
    ("cx", "csx"): {
        (0, 1): True,
        (0, None): True,
        (1, 0): False,
        (1, None): False,
        (None, 0): False,
        (None, 1): True,
    },
    ("cx", "cy"): {
        (0, 1): False,
        (0, None): True,
        (1, 0): False,
        (1, None): False,
        (None, 0): False,
        (None, 1): False,
    },
    ("cx", "cz"): {
        (0, 1): False,
        (0, None): True,
        (1, 0): False,
        (1, None): True,
        (None, 0): False,
        (None, 1): False,
    },
    ("cx", "ccz"): {
        (0, 1): False,
        (0, 2): False,
        (0, None): True,
        (1, 0): False,
        (1, 2): False,
        (1, None): True,
        (2, 0): False,
        (2, 1): False,
        (2, None): True,
        (None, 0): False,
        (None, 1): False,
        (None, 2): False,
    },
    ("cx", "rccx"): {
        (0, 1): False,
        (0, 2): False,
        (0, None): True,
        (1, 0): False,
        (1, 2): False,
        (1, None): True,
        (2, 0): False,
        (2, 1): False,
        (2, None): False,
        (None, 0): False,
        (None, 1): False,
        (None, 2): False,
    },
    ("cx", "rcccx"): {
        (0, 1): False,
        (0, 2): False,
        (0, 3): False,
        (0, None): True,
        (1, 0): False,
        (1, 2): False,
        (1, 3): False,
        (1, None): True,
        (2, 0): False,
        (2, 1): False,
        (2, 3): False,
        (2, None): True,
        (3, 0): False,
        (3, 1): False,
        (3, 2): False,
        (3, None): False,
        (None, 0): False,
        (None, 1): False,
        (None, 2): False,
        (None, 3): False,
    },
    ("cx", "ecr"): {
        (0, 1): False,
        (0, None): False,
        (1, 0): False,
        (1, None): False,
        (None, 0): False,
        (None, 1): True,
    },
    ("cx", "csdg"): {
        (0, 1): False,
        (0, None): True,
        (1, 0): False,
        (1, None): True,
        (None, 0): False,
        (None, 1): False,
    },
    ("cx", "swap"): False,
    ("cx", "iswap"): False,
    ("cx", "rxx"): {
        (0, 1): False,
        (0, None): False,
        (1, 0): False,
        (1, None): False,
        (None, 0): True,
        (None, 1): True,
    },
    ("cx", "ryy"): False,
    ("cx", "rzz"): {
        (0, 1): False,
        (0, None): True,
        (1, 0): False,
        (1, None): True,
        (None, 0): False,
        (None, 1): False,
    },
    ("cx", "rzx"): {
        (0, 1): True,
        (0, None): True,
        (1, 0): False,
        (1, None): False,
        (None, 0): False,
        (None, 1): True,
    },
    ("c3sx", "c3sx"): {
        (0, 1, 2, 3): True,
        (0, 1, 2, None): True,
        (0, 1, 3, 2): False,
        (0, 1, 3, None): False,
        (0, 1, None, 2): False,
        (0, 1, None, 3): True,
        (0, 1, None, None): True,
        (0, 2, 1, 3): True,
        (0, 2, 1, None): True,
        (0, 2, 3, 1): False,
        (0, 2, 3, None): False,
        (0, 2, None, 1): False,
        (0, 2, None, 3): True,
        (0, 2, None, None): True,
        (0, 3, 1, 2): False,
        (0, 3, 1, None): False,
        (0, 3, 2, 1): False,
        (0, 3, 2, None): False,
        (0, 3, None, 1): False,
        (0, 3, None, 2): False,
        (0, 3, None, None): False,
        (0, None, 1, 2): False,
        (0, None, 1, 3): True,
        (0, None, 1, None): True,
        (0, None, 2, 1): False,
        (0, None, 2, 3): True,
        (0, None, 2, None): True,
        (0, None, 3, 1): False,
        (0, None, 3, 2): False,
        (0, None, 3, None): False,
        (0, None, None, 1): False,
        (0, None, None, 2): False,
        (0, None, None, 3): True,
        (0, None, None, None): True,
        (1, 0, 2, 3): True,
        (1, 0, 2, None): True,
        (1, 0, 3, 2): False,
        (1, 0, 3, None): False,
        (1, 0, None, 2): False,
        (1, 0, None, 3): True,
        (1, 0, None, None): True,
        (1, 2, 0, 3): True,
        (1, 2, 0, None): True,
        (1, 2, 3, 0): False,
        (1, 2, 3, None): False,
        (1, 2, None, 0): False,
        (1, 2, None, 3): True,
        (1, 2, None, None): True,
        (1, 3, 0, 2): False,
        (1, 3, 0, None): False,
        (1, 3, 2, 0): False,
        (1, 3, 2, None): False,
        (1, 3, None, 0): False,
        (1, 3, None, 2): False,
        (1, 3, None, None): False,
        (1, None, 0, 2): False,
        (1, None, 0, 3): True,
        (1, None, 0, None): True,
        (1, None, 2, 0): False,
        (1, None, 2, 3): True,
        (1, None, 2, None): True,
        (1, None, 3, 0): False,
        (1, None, 3, 2): False,
        (1, None, 3, None): False,
        (1, None, None, 0): False,
        (1, None, None, 2): False,
        (1, None, None, 3): True,
        (1, None, None, None): True,
        (2, 0, 1, 3): True,
        (2, 0, 1, None): True,
        (2, 0, 3, 1): False,
        (2, 0, 3, None): False,
        (2, 0, None, 1): False,
        (2, 0, None, 3): True,
        (2, 0, None, None): True,
        (2, 1, 0, 3): True,
        (2, 1, 0, None): True,
        (2, 1, 3, 0): False,
        (2, 1, 3, None): False,
        (2, 1, None, 0): False,
        (2, 1, None, 3): True,
        (2, 1, None, None): True,
        (2, 3, 0, 1): False,
        (2, 3, 0, None): False,
        (2, 3, 1, 0): False,
        (2, 3, 1, None): False,
        (2, 3, None, 0): False,
        (2, 3, None, 1): False,
        (2, 3, None, None): False,
        (2, None, 0, 1): False,
        (2, None, 0, 3): True,
        (2, None, 0, None): True,
        (2, None, 1, 0): False,
        (2, None, 1, 3): True,
        (2, None, 1, None): True,
        (2, None, 3, 0): False,
        (2, None, 3, 1): False,
        (2, None, 3, None): False,
        (2, None, None, 0): False,
        (2, None, None, 1): False,
        (2, None, None, 3): True,
        (2, None, None, None): True,
        (3, 0, 1, 2): False,
        (3, 0, 1, None): False,
        (3, 0, 2, 1): False,
        (3, 0, 2, None): False,
        (3, 0, None, 1): False,
        (3, 0, None, 2): False,
        (3, 0, None, None): False,
        (3, 1, 0, 2): False,
        (3, 1, 0, None): False,
        (3, 1, 2, 0): False,
        (3, 1, 2, None): False,
        (3, 1, None, 0): False,
        (3, 1, None, 2): False,
        (3, 1, None, None): False,
        (3, 2, 0, 1): False,
        (3, 2, 0, None): False,
        (3, 2, 1, 0): False,
        (3, 2, 1, None): False,
        (3, 2, None, 0): False,
        (3, 2, None, 1): False,
        (3, 2, None, None): False,
        (3, None, 0, 1): False,
        (3, None, 0, 2): False,
        (3, None, 0, None): False,
        (3, None, 1, 0): False,
        (3, None, 1, 2): False,
        (3, None, 1, None): False,
        (3, None, 2, 0): False,
        (3, None, 2, 1): False,
        (3, None, 2, None): False,
        (3, None, None, 0): False,
        (3, None, None, 1): False,
        (3, None, None, 2): False,
        (3, None, None, None): False,
        (None, 0, 1, 2): False,
        (None, 0, 1, 3): True,
        (None, 0, 1, None): True,
        (None, 0, 2, 1): False,
        (None, 0, 2, 3): True,
        (None, 0, 2, None): True,
        (None, 0, 3, 1): False,
        (None, 0, 3, 2): False,
        (None, 0, 3, None): False,
        (None, 0, None, 1): False,
        (None, 0, None, 2): False,
        (None, 0, None, 3): True,
        (None, 0, None, None): True,
        (None, 1, 0, 2): False,
        (None, 1, 0, 3): True,
        (None, 1, 0, None): True,
        (None, 1, 2, 0): False,
        (None, 1, 2, 3): True,
        (None, 1, 2, None): True,
        (None, 1, 3, 0): False,
        (None, 1, 3, 2): False,
        (None, 1, 3, None): False,
        (None, 1, None, 0): False,
        (None, 1, None, 2): False,
        (None, 1, None, 3): True,
        (None, 1, None, None): True,
        (None, 2, 0, 1): False,
        (None, 2, 0, 3): True,
        (None, 2, 0, None): True,
        (None, 2, 1, 0): False,
        (None, 2, 1, 3): True,
        (None, 2, 1, None): True,
        (None, 2, 3, 0): False,
        (None, 2, 3, 1): False,
        (None, 2, 3, None): False,
        (None, 2, None, 0): False,
        (None, 2, None, 1): False,
        (None, 2, None, 3): True,
        (None, 2, None, None): True,
        (None, 3, 0, 1): False,
        (None, 3, 0, 2): False,
        (None, 3, 0, None): False,
        (None, 3, 1, 0): False,
        (None, 3, 1, 2): False,
        (None, 3, 1, None): False,
        (None, 3, 2, 0): False,
        (None, 3, 2, 1): False,
        (None, 3, 2, None): False,
        (None, 3, None, 0): False,
        (None, 3, None, 1): False,
        (None, 3, None, 2): False,
        (None, 3, None, None): False,
        (None, None, 0, 1): False,
        (None, None, 0, 2): False,
        (None, None, 0, 3): True,
        (None, None, 0, None): True,
        (None, None, 1, 0): False,
        (None, None, 1, 2): False,
        (None, None, 1, 3): True,
        (None, None, 1, None): True,
        (None, None, 2, 0): False,
        (None, None, 2, 1): False,
        (None, None, 2, 3): True,
        (None, None, 2, None): True,
        (None, None, 3, 0): False,
        (None, None, 3, 1): False,
        (None, None, 3, 2): False,
        (None, None, 3, None): False,
        (None, None, None, 0): False,
        (None, None, None, 1): False,
        (None, None, None, 2): False,
        (None, None, None, 3): True,
    },
    ("c3sx", "rcccx"): {
        (0, 1, 2, 3): False,
        (0, 1, 2, None): True,
        (0, 1, 3, 2): False,
        (0, 1, 3, None): False,
        (0, 1, None, 2): False,
        (0, 1, None, 3): False,
        (0, 1, None, None): True,
        (0, 2, 1, 3): False,
        (0, 2, 1, None): True,
        (0, 2, 3, 1): False,
        (0, 2, 3, None): False,
        (0, 2, None, 1): False,
        (0, 2, None, 3): False,
        (0, 2, None, None): True,
        (0, 3, 1, 2): False,
        (0, 3, 1, None): False,
        (0, 3, 2, 1): False,
        (0, 3, 2, None): False,
        (0, 3, None, 1): False,
        (0, 3, None, 2): False,
        (0, 3, None, None): False,
        (0, None, 1, 2): False,
        (0, None, 1, 3): False,
        (0, None, 1, None): True,
        (0, None, 2, 1): False,
        (0, None, 2, 3): False,
        (0, None, 2, None): True,
        (0, None, 3, 1): False,
        (0, None, 3, 2): False,
        (0, None, 3, None): False,
        (0, None, None, 1): False,
        (0, None, None, 2): False,
        (0, None, None, 3): False,
        (0, None, None, None): True,
        (1, 0, 2, 3): False,
        (1, 0, 2, None): True,
        (1, 0, 3, 2): False,
        (1, 0, 3, None): False,
        (1, 0, None, 2): False,
        (1, 0, None, 3): False,
        (1, 0, None, None): True,
        (1, 2, 0, 3): False,
        (1, 2, 0, None): True,
        (1, 2, 3, 0): False,
        (1, 2, 3, None): False,
        (1, 2, None, 0): False,
        (1, 2, None, 3): False,
        (1, 2, None, None): True,
        (1, 3, 0, 2): False,
        (1, 3, 0, None): False,
        (1, 3, 2, 0): False,
        (1, 3, 2, None): False,
        (1, 3, None, 0): False,
        (1, 3, None, 2): False,
        (1, 3, None, None): False,
        (1, None, 0, 2): False,
        (1, None, 0, 3): False,
        (1, None, 0, None): True,
        (1, None, 2, 0): False,
        (1, None, 2, 3): False,
        (1, None, 2, None): True,
        (1, None, 3, 0): False,
        (1, None, 3, 2): False,
        (1, None, 3, None): False,
        (1, None, None, 0): False,
        (1, None, None, 2): False,
        (1, None, None, 3): False,
        (1, None, None, None): True,
        (2, 0, 1, 3): False,
        (2, 0, 1, None): True,
        (2, 0, 3, 1): False,
        (2, 0, 3, None): False,
        (2, 0, None, 1): False,
        (2, 0, None, 3): False,
        (2, 0, None, None): True,
        (2, 1, 0, 3): False,
        (2, 1, 0, None): True,
        (2, 1, 3, 0): False,
        (2, 1, 3, None): False,
        (2, 1, None, 0): False,
        (2, 1, None, 3): False,
        (2, 1, None, None): True,
        (2, 3, 0, 1): False,
        (2, 3, 0, None): False,
        (2, 3, 1, 0): False,
        (2, 3, 1, None): False,
        (2, 3, None, 0): False,
        (2, 3, None, 1): False,
        (2, 3, None, None): False,
        (2, None, 0, 1): False,
        (2, None, 0, 3): False,
        (2, None, 0, None): True,
        (2, None, 1, 0): False,
        (2, None, 1, 3): False,
        (2, None, 1, None): True,
        (2, None, 3, 0): False,
        (2, None, 3, 1): False,
        (2, None, 3, None): False,
        (2, None, None, 0): False,
        (2, None, None, 1): False,
        (2, None, None, 3): False,
        (2, None, None, None): True,
        (3, 0, 1, 2): False,
        (3, 0, 1, None): False,
        (3, 0, 2, 1): False,
        (3, 0, 2, None): False,
        (3, 0, None, 1): False,
        (3, 0, None, 2): False,
        (3, 0, None, None): False,
        (3, 1, 0, 2): False,
        (3, 1, 0, None): False,
        (3, 1, 2, 0): False,
        (3, 1, 2, None): False,
        (3, 1, None, 0): False,
        (3, 1, None, 2): False,
        (3, 1, None, None): False,
        (3, 2, 0, 1): False,
        (3, 2, 0, None): False,
        (3, 2, 1, 0): False,
        (3, 2, 1, None): False,
        (3, 2, None, 0): False,
        (3, 2, None, 1): False,
        (3, 2, None, None): False,
        (3, None, 0, 1): False,
        (3, None, 0, 2): False,
        (3, None, 0, None): False,
        (3, None, 1, 0): False,
        (3, None, 1, 2): False,
        (3, None, 1, None): False,
        (3, None, 2, 0): False,
        (3, None, 2, 1): False,
        (3, None, 2, None): False,
        (3, None, None, 0): False,
        (3, None, None, 1): False,
        (3, None, None, 2): False,
        (3, None, None, None): False,
        (None, 0, 1, 2): False,
        (None, 0, 1, 3): False,
        (None, 0, 1, None): True,
        (None, 0, 2, 1): False,
        (None, 0, 2, 3): False,
        (None, 0, 2, None): True,
        (None, 0, 3, 1): False,
        (None, 0, 3, 2): False,
        (None, 0, 3, None): False,
        (None, 0, None, 1): False,
        (None, 0, None, 2): False,
        (None, 0, None, 3): False,
        (None, 0, None, None): True,
        (None, 1, 0, 2): False,
        (None, 1, 0, 3): False,
        (None, 1, 0, None): True,
        (None, 1, 2, 0): False,
        (None, 1, 2, 3): False,
        (None, 1, 2, None): True,
        (None, 1, 3, 0): False,
        (None, 1, 3, 2): False,
        (None, 1, 3, None): False,
        (None, 1, None, 0): False,
        (None, 1, None, 2): False,
        (None, 1, None, 3): False,
        (None, 1, None, None): True,
        (None, 2, 0, 1): False,
        (None, 2, 0, 3): False,
        (None, 2, 0, None): True,
        (None, 2, 1, 0): False,
        (None, 2, 1, 3): False,
        (None, 2, 1, None): True,
        (None, 2, 3, 0): False,
        (None, 2, 3, 1): False,
        (None, 2, 3, None): False,
        (None, 2, None, 0): False,
        (None, 2, None, 1): False,
        (None, 2, None, 3): False,
        (None, 2, None, None): True,
        (None, 3, 0, 1): False,
        (None, 3, 0, 2): False,
        (None, 3, 0, None): False,
        (None, 3, 1, 0): False,
        (None, 3, 1, 2): False,
        (None, 3, 1, None): False,
        (None, 3, 2, 0): False,
        (None, 3, 2, 1): False,
        (None, 3, 2, None): False,
        (None, 3, None, 0): False,
        (None, 3, None, 1): False,
        (None, 3, None, 2): False,
        (None, 3, None, None): False,
        (None, None, 0, 1): False,
        (None, None, 0, 2): False,
        (None, None, 0, 3): False,
        (None, None, 0, None): True,
        (None, None, 1, 0): False,
        (None, None, 1, 2): False,
        (None, None, 1, 3): False,
        (None, None, 1, None): True,
        (None, None, 2, 0): False,
        (None, None, 2, 1): False,
        (None, None, 2, 3): False,
        (None, None, 2, None): True,
        (None, None, 3, 0): False,
        (None, None, 3, 1): False,
        (None, None, 3, 2): False,
        (None, None, 3, None): False,
        (None, None, None, 0): False,
        (None, None, None, 1): False,
        (None, None, None, 2): False,
        (None, None, None, 3): False,
    },
    ("ccx", "c3sx"): {
        (0, 1, 2): False,
        (0, 1, 3): True,
        (0, 1, None): True,
        (0, 2, 1): False,
        (0, 2, 3): True,
        (0, 2, None): True,
        (0, 3, 1): False,
        (0, 3, 2): False,
        (0, 3, None): False,
        (0, None, 1): False,
        (0, None, 2): False,
        (0, None, 3): True,
        (0, None, None): True,
        (1, 0, 2): False,
        (1, 0, 3): True,
        (1, 0, None): True,
        (1, 2, 0): False,
        (1, 2, 3): True,
        (1, 2, None): True,
        (1, 3, 0): False,
        (1, 3, 2): False,
        (1, 3, None): False,
        (1, None, 0): False,
        (1, None, 2): False,
        (1, None, 3): True,
        (1, None, None): True,
        (2, 0, 1): False,
        (2, 0, 3): True,
        (2, 0, None): True,
        (2, 1, 0): False,
        (2, 1, 3): True,
        (2, 1, None): True,
        (2, 3, 0): False,
        (2, 3, 1): False,
        (2, 3, None): False,
        (2, None, 0): False,
        (2, None, 1): False,
        (2, None, 3): True,
        (2, None, None): True,
        (3, 0, 1): False,
        (3, 0, 2): False,
        (3, 0, None): False,
        (3, 1, 0): False,
        (3, 1, 2): False,
        (3, 1, None): False,
        (3, 2, 0): False,
        (3, 2, 1): False,
        (3, 2, None): False,
        (3, None, 0): False,
        (3, None, 1): False,
        (3, None, 2): False,
        (3, None, None): False,
        (None, 0, 1): False,
        (None, 0, 2): False,
        (None, 0, 3): True,
        (None, 0, None): True,
        (None, 1, 0): False,
        (None, 1, 2): False,
        (None, 1, 3): True,
        (None, 1, None): True,
        (None, 2, 0): False,
        (None, 2, 1): False,
        (None, 2, 3): True,
        (None, 2, None): True,
        (None, 3, 0): False,
        (None, 3, 1): False,
        (None, 3, 2): False,
        (None, 3, None): False,
        (None, None, 0): False,
        (None, None, 1): False,
        (None, None, 2): False,
        (None, None, 3): True,
    },
    ("ccx", "ccx"): {
        (0, 1, 2): True,
        (0, 1, None): True,
        (0, 2, 1): False,
        (0, 2, None): False,
        (0, None, 1): False,
        (0, None, 2): True,
        (0, None, None): True,
        (1, 0, 2): True,
        (1, 0, None): True,
        (1, 2, 0): False,
        (1, 2, None): False,
        (1, None, 0): False,
        (1, None, 2): True,
        (1, None, None): True,
        (2, 0, 1): False,
        (2, 0, None): False,
        (2, 1, 0): False,
        (2, 1, None): False,
        (2, None, 0): False,
        (2, None, 1): False,
        (2, None, None): False,
        (None, 0, 1): False,
        (None, 0, 2): True,
        (None, 0, None): True,
        (None, 1, 0): False,
        (None, 1, 2): True,
        (None, 1, None): True,
        (None, 2, 0): False,
        (None, 2, 1): False,
        (None, 2, None): False,
        (None, None, 0): False,
        (None, None, 1): False,
        (None, None, 2): True,
    },
    ("ccx", "cswap"): {
        (0, 1, 2): False,
        (0, 1, None): False,
        (0, 2, 1): False,
        (0, 2, None): False,
        (0, None, 1): False,
        (0, None, 2): False,
        (0, None, None): True,
        (1, 0, 2): False,
        (1, 0, None): False,
        (1, 2, 0): True,
        (1, 2, None): True,
        (1, None, 0): False,
        (1, None, 2): False,
        (1, None, None): False,
        (2, 0, 1): False,
        (2, 0, None): False,
        (2, 1, 0): True,
        (2, 1, None): True,
        (2, None, 0): False,
        (2, None, 1): False,
        (2, None, None): False,
        (None, 0, 1): False,
        (None, 0, 2): False,
        (None, 0, None): True,
        (None, 1, 0): False,
        (None, 1, 2): False,
        (None, 1, None): False,
        (None, 2, 0): False,
        (None, 2, 1): False,
        (None, 2, None): False,
        (None, None, 0): False,
        (None, None, 1): False,
        (None, None, 2): False,
    },
    ("ccx", "ccz"): {
        (0, 1, 2): False,
        (0, 1, None): True,
        (0, 2, 1): False,
        (0, 2, None): True,
        (0, None, 1): False,
        (0, None, 2): False,
        (0, None, None): True,
        (1, 0, 2): False,
        (1, 0, None): True,
        (1, 2, 0): False,
        (1, 2, None): True,
        (1, None, 0): False,
        (1, None, 2): False,
        (1, None, None): True,
        (2, 0, 1): False,
        (2, 0, None): True,
        (2, 1, 0): False,
        (2, 1, None): True,
        (2, None, 0): False,
        (2, None, 1): False,
        (2, None, None): True,
        (None, 0, 1): False,
        (None, 0, 2): False,
        (None, 0, None): True,
        (None, 1, 0): False,
        (None, 1, 2): False,
        (None, 1, None): True,
        (None, 2, 0): False,
        (None, 2, 1): False,
        (None, 2, None): True,
        (None, None, 0): False,
        (None, None, 1): False,
        (None, None, 2): False,
    },
    ("ccx", "rccx"): {
        (0, 1, 2): False,
        (0, 1, None): True,
        (0, 2, 1): False,
        (0, 2, None): False,
        (0, None, 1): False,
        (0, None, 2): False,
        (0, None, None): True,
        (1, 0, 2): False,
        (1, 0, None): True,
        (1, 2, 0): False,
        (1, 2, None): False,
        (1, None, 0): False,
        (1, None, 2): False,
        (1, None, None): True,
        (2, 0, 1): False,
        (2, 0, None): False,
        (2, 1, 0): False,
        (2, 1, None): False,
        (2, None, 0): False,
        (2, None, 1): False,
        (2, None, None): False,
        (None, 0, 1): False,
        (None, 0, 2): False,
        (None, 0, None): True,
        (None, 1, 0): False,
        (None, 1, 2): False,
        (None, 1, None): True,
        (None, 2, 0): False,
        (None, 2, 1): False,
        (None, 2, None): False,
        (None, None, 0): False,
        (None, None, 1): False,
        (None, None, 2): False,
    },
    ("ccx", "rcccx"): {
        (0, 1, 2): False,
        (0, 1, 3): False,
        (0, 1, None): True,
        (0, 2, 1): False,
        (0, 2, 3): False,
        (0, 2, None): True,
        (0, 3, 1): False,
        (0, 3, 2): False,
        (0, 3, None): False,
        (0, None, 1): False,
        (0, None, 2): False,
        (0, None, 3): False,
        (0, None, None): True,
        (1, 0, 2): False,
        (1, 0, 3): False,
        (1, 0, None): True,
        (1, 2, 0): False,
        (1, 2, 3): False,
        (1, 2, None): True,
        (1, 3, 0): False,
        (1, 3, 2): False,
        (1, 3, None): False,
        (1, None, 0): False,
        (1, None, 2): False,
        (1, None, 3): False,
        (1, None, None): True,
        (2, 0, 1): False,
        (2, 0, 3): False,
        (2, 0, None): True,
        (2, 1, 0): False,
        (2, 1, 3): False,
        (2, 1, None): True,
        (2, 3, 0): False,
        (2, 3, 1): False,
        (2, 3, None): False,
        (2, None, 0): False,
        (2, None, 1): False,
        (2, None, 3): False,
        (2, None, None): True,
        (3, 0, 1): False,
        (3, 0, 2): False,
        (3, 0, None): False,
        (3, 1, 0): False,
        (3, 1, 2): False,
        (3, 1, None): False,
        (3, 2, 0): False,
        (3, 2, 1): False,
        (3, 2, None): False,
        (3, None, 0): False,
        (3, None, 1): False,
        (3, None, 2): False,
        (3, None, None): False,
        (None, 0, 1): False,
        (None, 0, 2): False,
        (None, 0, 3): False,
        (None, 0, None): True,
        (None, 1, 0): False,
        (None, 1, 2): False,
        (None, 1, 3): False,
        (None, 1, None): True,
        (None, 2, 0): False,
        (None, 2, 1): False,
        (None, 2, 3): False,
        (None, 2, None): True,
        (None, 3, 0): False,
        (None, 3, 1): False,
        (None, 3, 2): False,
        (None, 3, None): False,
        (None, None, 0): False,
        (None, None, 1): False,
        (None, None, 2): False,
        (None, None, 3): False,
    },
    ("dcx", "c3sx"): False,
    ("dcx", "ccx"): False,
    ("dcx", "dcx"): {
        (0, 1): True,
        (0, None): False,
        (1, 0): True,
        (1, None): False,
        (None, 0): False,
        (None, 1): False,
    },
    ("dcx", "cswap"): False,
    ("dcx", "ccz"): False,
    ("dcx", "rccx"): False,
    ("dcx", "rcccx"): False,
    ("dcx", "ecr"): False,
    ("dcx", "csdg"): False,
    ("dcx", "swap"): False,
    ("dcx", "iswap"): False,
    ("dcx", "rxx"): False,
    ("dcx", "ryy"): False,
    ("dcx", "rzz"): False,
    ("dcx", "rzx"): False,
    ("ch", "cx"): {
        (0, 1): False,
        (0, None): True,
        (1, 0): False,
        (1, None): False,
        (None, 0): False,
        (None, 1): False,
    },
    ("ch", "c3sx"): {
        (0, 1): False,
        (0, 2): False,
        (0, 3): False,
        (0, None): True,
        (1, 0): False,
        (1, 2): False,
        (1, 3): False,
        (1, None): True,
        (2, 0): False,
        (2, 1): False,
        (2, 3): False,
        (2, None): True,
        (3, 0): False,
        (3, 1): False,
        (3, 2): False,
        (3, None): False,
        (None, 0): False,
        (None, 1): False,
        (None, 2): False,
        (None, 3): False,
    },
    ("ch", "ccx"): {
        (0, 1): False,
        (0, 2): False,
        (0, None): True,
        (1, 0): False,
        (1, 2): False,
        (1, None): True,
        (2, 0): False,
        (2, 1): False,
        (2, None): False,
        (None, 0): False,
        (None, 1): False,
        (None, 2): False,
    },
    ("ch", "dcx"): False,
    ("ch", "ch"): {
        (0, 1): True,
        (0, None): True,
        (1, 0): False,
        (1, None): False,
        (None, 0): False,
        (None, 1): True,
    },
    ("ch", "cswap"): {
        (0, 1): False,
        (0, 2): False,
        (0, None): True,
        (1, 0): False,
        (1, 2): False,
        (1, None): False,
        (2, 0): False,
        (2, 1): False,
        (2, None): False,
        (None, 0): False,
        (None, 1): False,
        (None, 2): False,
    },
    ("ch", "csx"): {
        (0, 1): False,
        (0, None): True,
        (1, 0): False,
        (1, None): False,
        (None, 0): False,
        (None, 1): False,
    },
    ("ch", "cy"): {
        (0, 1): False,
        (0, None): True,
        (1, 0): False,
        (1, None): False,
        (None, 0): False,
        (None, 1): False,
    },
    ("ch", "cz"): {
        (0, 1): False,
        (0, None): True,
        (1, 0): False,
        (1, None): True,
        (None, 0): False,
        (None, 1): False,
    },
    ("ch", "ccz"): {
        (0, 1): False,
        (0, 2): False,
        (0, None): True,
        (1, 0): False,
        (1, 2): False,
        (1, None): True,
        (2, 0): False,
        (2, 1): False,
        (2, None): True,
        (None, 0): False,
        (None, 1): False,
        (None, 2): False,
    },
    ("ch", "rccx"): {
        (0, 1): False,
        (0, 2): False,
        (0, None): True,
        (1, 0): False,
        (1, 2): False,
        (1, None): True,
        (2, 0): False,
        (2, 1): False,
        (2, None): False,
        (None, 0): False,
        (None, 1): False,
        (None, 2): False,
    },
    ("ch", "rcccx"): {
        (0, 1): False,
        (0, 2): False,
        (0, 3): False,
        (0, None): True,
        (1, 0): False,
        (1, 2): False,
        (1, 3): False,
        (1, None): True,
        (2, 0): False,
        (2, 1): False,
        (2, 3): False,
        (2, None): True,
        (3, 0): False,
        (3, 1): False,
        (3, 2): False,
        (3, None): False,
        (None, 0): False,
        (None, 1): False,
        (None, 2): False,
        (None, 3): False,
    },
    ("ch", "ecr"): False,
    ("ch", "cs"): {
        (0, 1): False,
        (0, None): True,
        (1, 0): False,
        (1, None): True,
        (None, 0): False,
        (None, 1): False,
    },
    ("ch", "csdg"): {
        (0, 1): False,
        (0, None): True,
        (1, 0): False,
        (1, None): True,
        (None, 0): False,
        (None, 1): False,
    },
    ("ch", "swap"): False,
    ("ch", "iswap"): False,
    ("ch", "rxx"): False,
    ("ch", "ryy"): False,
    ("ch", "rzz"): {
        (0, 1): False,
        (0, None): True,
        (1, 0): False,
        (1, None): True,
        (None, 0): False,
        (None, 1): False,
    },
    ("ch", "rzx"): {
        (0, 1): False,
        (0, None): True,
        (1, 0): False,
        (1, None): False,
        (None, 0): False,
        (None, 1): False,
    },
    ("cswap", "c3sx"): {
        (0, 1, 2): True,
        (0, 1, 3): False,
        (0, 1, None): False,
        (0, 2, 1): True,
        (0, 2, 3): False,
        (0, 2, None): False,
        (0, 3, 1): False,
        (0, 3, 2): False,
        (0, 3, None): False,
        (0, None, 1): False,
        (0, None, 2): False,
        (0, None, 3): False,
        (0, None, None): True,
        (1, 0, 2): True,
        (1, 0, 3): False,
        (1, 0, None): False,
        (1, 2, 0): True,
        (1, 2, 3): False,
        (1, 2, None): False,
        (1, 3, 0): False,
        (1, 3, 2): False,
        (1, 3, None): False,
        (1, None, 0): False,
        (1, None, 2): False,
        (1, None, 3): False,
        (1, None, None): True,
        (2, 0, 1): True,
        (2, 0, 3): False,
        (2, 0, None): False,
        (2, 1, 0): True,
        (2, 1, 3): False,
        (2, 1, None): False,
        (2, 3, 0): False,
        (2, 3, 1): False,
        (2, 3, None): False,
        (2, None, 0): False,
        (2, None, 1): False,
        (2, None, 3): False,
        (2, None, None): True,
        (3, 0, 1): True,
        (3, 0, 2): True,
        (3, 0, None): False,
        (3, 1, 0): True,
        (3, 1, 2): True,
        (3, 1, None): False,
        (3, 2, 0): True,
        (3, 2, 1): True,
        (3, 2, None): False,
        (3, None, 0): False,
        (3, None, 1): False,
        (3, None, 2): False,
        (3, None, None): False,
        (None, 0, 1): True,
        (None, 0, 2): True,
        (None, 0, 3): False,
        (None, 0, None): False,
        (None, 1, 0): True,
        (None, 1, 2): True,
        (None, 1, 3): False,
        (None, 1, None): False,
        (None, 2, 0): True,
        (None, 2, 1): True,
        (None, 2, 3): False,
        (None, 2, None): False,
        (None, 3, 0): False,
        (None, 3, 1): False,
        (None, 3, 2): False,
        (None, 3, None): False,
        (None, None, 0): False,
        (None, None, 1): False,
        (None, None, 2): False,
        (None, None, 3): False,
    },
    ("cswap", "cswap"): {
        (0, 1, 2): True,
        (0, 1, None): False,
        (0, 2, 1): True,
        (0, 2, None): False,
        (0, None, 1): False,
        (0, None, 2): False,
        (0, None, None): True,
        (1, 0, 2): False,
        (1, 0, None): False,
        (1, 2, 0): False,
        (1, 2, None): False,
        (1, None, 0): False,
        (1, None, 2): False,
        (1, None, None): False,
        (2, 0, 1): False,
        (2, 0, None): False,
        (2, 1, 0): False,
        (2, 1, None): False,
        (2, None, 0): False,
        (2, None, 1): False,
        (2, None, None): False,
        (None, 0, 1): False,
        (None, 0, 2): False,
        (None, 0, None): False,
        (None, 1, 0): False,
        (None, 1, 2): True,
        (None, 1, None): False,
        (None, 2, 0): False,
        (None, 2, 1): True,
        (None, 2, None): False,
        (None, None, 0): False,
        (None, None, 1): False,
        (None, None, 2): False,
    },
    ("cswap", "rcccx"): {
        (0, 1, 2): False,
        (0, 1, 3): False,
        (0, 1, None): False,
        (0, 2, 1): False,
        (0, 2, 3): False,
        (0, 2, None): False,
        (0, 3, 1): False,
        (0, 3, 2): False,
        (0, 3, None): False,
        (0, None, 1): False,
        (0, None, 2): False,
        (0, None, 3): False,
        (0, None, None): True,
        (1, 0, 2): False,
        (1, 0, 3): False,
        (1, 0, None): False,
        (1, 2, 0): False,
        (1, 2, 3): False,
        (1, 2, None): False,
        (1, 3, 0): False,
        (1, 3, 2): False,
        (1, 3, None): False,
        (1, None, 0): False,
        (1, None, 2): False,
        (1, None, 3): False,
        (1, None, None): True,
        (2, 0, 1): True,
        (2, 0, 3): False,
        (2, 0, None): False,
        (2, 1, 0): True,
        (2, 1, 3): False,
        (2, 1, None): False,
        (2, 3, 0): False,
        (2, 3, 1): False,
        (2, 3, None): False,
        (2, None, 0): False,
        (2, None, 1): False,
        (2, None, 3): False,
        (2, None, None): True,
        (3, 0, 1): True,
        (3, 0, 2): False,
        (3, 0, None): False,
        (3, 1, 0): True,
        (3, 1, 2): False,
        (3, 1, None): False,
        (3, 2, 0): False,
        (3, 2, 1): False,
        (3, 2, None): False,
        (3, None, 0): False,
        (3, None, 1): False,
        (3, None, 2): False,
        (3, None, None): False,
        (None, 0, 1): True,
        (None, 0, 2): False,
        (None, 0, 3): False,
        (None, 0, None): False,
        (None, 1, 0): True,
        (None, 1, 2): False,
        (None, 1, 3): False,
        (None, 1, None): False,
        (None, 2, 0): False,
        (None, 2, 1): False,
        (None, 2, 3): False,
        (None, 2, None): False,
        (None, 3, 0): False,
        (None, 3, 1): False,
        (None, 3, 2): False,
        (None, 3, None): False,
        (None, None, 0): False,
        (None, None, 1): False,
        (None, None, 2): False,
        (None, None, 3): False,
    },
    ("csx", "c3sx"): {
        (0, 1): False,
        (0, 2): False,
        (0, 3): True,
        (0, None): True,
        (1, 0): False,
        (1, 2): False,
        (1, 3): True,
        (1, None): True,
        (2, 0): False,
        (2, 1): False,
        (2, 3): True,
        (2, None): True,
        (3, 0): False,
        (3, 1): False,
        (3, 2): False,
        (3, None): False,
        (None, 0): False,
        (None, 1): False,
        (None, 2): False,
        (None, 3): True,
    },
    ("csx", "ccx"): {
        (0, 1): False,
        (0, 2): True,
        (0, None): True,
        (1, 0): False,
        (1, 2): True,
        (1, None): True,
        (2, 0): False,
        (2, 1): False,
        (2, None): False,
        (None, 0): False,
        (None, 1): False,
        (None, 2): True,
    },
    ("csx", "dcx"): False,
    ("csx", "cswap"): {
        (0, 1): False,
        (0, 2): False,
        (0, None): True,
        (1, 0): False,
        (1, 2): False,
        (1, None): False,
        (2, 0): False,
        (2, 1): False,
        (2, None): False,
        (None, 0): False,
        (None, 1): False,
        (None, 2): False,
    },
    ("csx", "csx"): {
        (0, 1): True,
        (0, None): True,
        (1, 0): False,
        (1, None): False,
        (None, 0): False,
        (None, 1): True,
    },
    ("csx", "ccz"): {
        (0, 1): False,
        (0, 2): False,
        (0, None): True,
        (1, 0): False,
        (1, 2): False,
        (1, None): True,
        (2, 0): False,
        (2, 1): False,
        (2, None): True,
        (None, 0): False,
        (None, 1): False,
        (None, 2): False,
    },
    ("csx", "rccx"): {
        (0, 1): False,
        (0, 2): False,
        (0, None): True,
        (1, 0): False,
        (1, 2): False,
        (1, None): True,
        (2, 0): False,
        (2, 1): False,
        (2, None): False,
        (None, 0): False,
        (None, 1): False,
        (None, 2): False,
    },
    ("csx", "rcccx"): {
        (0, 1): False,
        (0, 2): False,
        (0, 3): False,
        (0, None): True,
        (1, 0): False,
        (1, 2): False,
        (1, 3): False,
        (1, None): True,
        (2, 0): False,
        (2, 1): False,
        (2, 3): False,
        (2, None): True,
        (3, 0): False,
        (3, 1): False,
        (3, 2): False,
        (3, None): False,
        (None, 0): False,
        (None, 1): False,
        (None, 2): False,
        (None, 3): False,
    },
    ("csx", "ecr"): {
        (0, 1): False,
        (0, None): False,
        (1, 0): False,
        (1, None): False,
        (None, 0): False,
        (None, 1): True,
    },
    ("csx", "csdg"): {
        (0, 1): False,
        (0, None): True,
        (1, 0): False,
        (1, None): True,
        (None, 0): False,
        (None, 1): False,
    },
    ("csx", "swap"): False,
    ("csx", "iswap"): False,
    ("csx", "rxx"): {
        (0, 1): False,
        (0, None): False,
        (1, 0): False,
        (1, None): False,
        (None, 0): True,
        (None, 1): True,
    },
    ("csx", "ryy"): False,
    ("csx", "rzz"): {
        (0, 1): False,
        (0, None): True,
        (1, 0): False,
        (1, None): True,
        (None, 0): False,
        (None, 1): False,
    },
    ("csx", "rzx"): {
        (0, 1): True,
        (0, None): True,
        (1, 0): False,
        (1, None): False,
        (None, 0): False,
        (None, 1): True,
    },
    ("cy", "c3sx"): {
        (0, 1): False,
        (0, 2): False,
        (0, 3): False,
        (0, None): True,
        (1, 0): False,
        (1, 2): False,
        (1, 3): False,
        (1, None): True,
        (2, 0): False,
        (2, 1): False,
        (2, 3): False,
        (2, None): True,
        (3, 0): False,
        (3, 1): False,
        (3, 2): False,
        (3, None): False,
        (None, 0): False,
        (None, 1): False,
        (None, 2): False,
        (None, 3): False,
    },
    ("cy", "ccx"): {
        (0, 1): False,
        (0, 2): False,
        (0, None): True,
        (1, 0): False,
        (1, 2): False,
        (1, None): True,
        (2, 0): False,
        (2, 1): False,
        (2, None): False,
        (None, 0): False,
        (None, 1): False,
        (None, 2): False,
    },
    ("cy", "dcx"): False,
    ("cy", "cswap"): {
        (0, 1): False,
        (0, 2): False,
        (0, None): True,
        (1, 0): False,
        (1, 2): False,
        (1, None): False,
        (2, 0): False,
        (2, 1): False,
        (2, None): False,
        (None, 0): False,
        (None, 1): False,
        (None, 2): False,
    },
    ("cy", "csx"): {
        (0, 1): False,
        (0, None): True,
        (1, 0): False,
        (1, None): False,
        (None, 0): False,
        (None, 1): False,
    },
    ("cy", "cy"): {
        (0, 1): True,
        (0, None): True,
        (1, 0): False,
        (1, None): False,
        (None, 0): False,
        (None, 1): True,
    },
    ("cy", "cz"): {
        (0, 1): False,
        (0, None): True,
        (1, 0): False,
        (1, None): True,
        (None, 0): False,
        (None, 1): False,
    },
    ("cy", "ccz"): {
        (0, 1): False,
        (0, 2): False,
        (0, None): True,
        (1, 0): False,
        (1, 2): False,
        (1, None): True,
        (2, 0): False,
        (2, 1): False,
        (2, None): True,
        (None, 0): False,
        (None, 1): False,
        (None, 2): False,
    },
    ("cy", "rccx"): {
        (0, 1): False,
        (0, 2): False,
        (0, None): True,
        (1, 0): False,
        (1, 2): True,
        (1, None): True,
        (2, 0): False,
        (2, 1): False,
        (2, None): False,
        (None, 0): False,
        (None, 1): False,
        (None, 2): False,
    },
    ("cy", "rcccx"): {
        (0, 1): False,
        (0, 2): False,
        (0, 3): False,
        (0, None): True,
        (1, 0): False,
        (1, 2): False,
        (1, 3): False,
        (1, None): True,
        (2, 0): False,
        (2, 1): False,
        (2, 3): True,
        (2, None): True,
        (3, 0): False,
        (3, 1): False,
        (3, 2): False,
        (3, None): False,
        (None, 0): False,
        (None, 1): False,
        (None, 2): False,
        (None, 3): False,
    },
    ("cy", "ecr"): False,
    ("cy", "csdg"): {
        (0, 1): False,
        (0, None): True,
        (1, 0): False,
        (1, None): True,
        (None, 0): False,
        (None, 1): False,
    },
    ("cy", "swap"): False,
    ("cy", "iswap"): False,
    ("cy", "rxx"): False,
    ("cy", "ryy"): {
        (0, 1): False,
        (0, None): False,
        (1, 0): False,
        (1, None): False,
        (None, 0): True,
        (None, 1): True,
    },
    ("cy", "rzz"): {
        (0, 1): False,
        (0, None): True,
        (1, 0): False,
        (1, None): True,
        (None, 0): False,
        (None, 1): False,
    },
    ("cy", "rzx"): {
        (0, 1): False,
        (0, None): True,
        (1, 0): False,
        (1, None): False,
        (None, 0): False,
        (None, 1): False,
    },
    ("cz", "c3sx"): {
        (0, 1): True,
        (0, 2): True,
        (0, 3): False,
        (0, None): True,
        (1, 0): True,
        (1, 2): True,
        (1, 3): False,
        (1, None): True,
        (2, 0): True,
        (2, 1): True,
        (2, 3): False,
        (2, None): True,
        (3, 0): False,
        (3, 1): False,
        (3, 2): False,
        (3, None): False,
        (None, 0): True,
        (None, 1): True,
        (None, 2): True,
        (None, 3): False,
    },
    ("cz", "ccx"): {
        (0, 1): True,
        (0, 2): False,
        (0, None): True,
        (1, 0): True,
        (1, 2): False,
        (1, None): True,
        (2, 0): False,
        (2, 1): False,
        (2, None): False,
        (None, 0): True,
        (None, 1): True,
        (None, 2): False,
    },
    ("cz", "dcx"): False,
    ("cz", "cswap"): {
        (0, 1): False,
        (0, 2): False,
        (0, None): True,
        (1, 0): False,
        (1, 2): True,
        (1, None): False,
        (2, 0): False,
        (2, 1): True,
        (2, None): False,
        (None, 0): True,
        (None, 1): False,
        (None, 2): False,
    },
    ("cz", "csx"): {
        (0, 1): False,
        (0, None): True,
        (1, 0): False,
        (1, None): False,
        (None, 0): True,
        (None, 1): False,
    },
    ("cz", "cz"): True,
    ("cz", "ccz"): True,
    ("cz", "rccx"): {
        (0, 1): True,
        (0, 2): False,
        (0, None): True,
        (1, 0): True,
        (1, 2): False,
        (1, None): True,
        (2, 0): False,
        (2, 1): False,
        (2, None): False,
        (None, 0): True,
        (None, 1): True,
        (None, 2): False,
    },
    ("cz", "rcccx"): {
        (0, 1): True,
        (0, 2): True,
        (0, 3): False,
        (0, None): True,
        (1, 0): True,
        (1, 2): True,
        (1, 3): False,
        (1, None): True,
        (2, 0): True,
        (2, 1): True,
        (2, 3): False,
        (2, None): True,
        (3, 0): False,
        (3, 1): False,
        (3, 2): False,
        (3, None): False,
        (None, 0): True,
        (None, 1): True,
        (None, 2): True,
        (None, 3): False,
    },
    ("cz", "ecr"): False,
    ("cz", "csdg"): True,
    ("cz", "swap"): {
        (0, 1): True,
        (0, None): False,
        (1, 0): True,
        (1, None): False,
        (None, 0): False,
        (None, 1): False,
    },
    ("cz", "iswap"): {
        (0, 1): True,
        (0, None): False,
        (1, 0): True,
        (1, None): False,
        (None, 0): False,
        (None, 1): False,
    },
    ("cz", "rxx"): False,
    ("cz", "ryy"): False,
    ("cz", "rzz"): True,
    ("cz", "rzx"): {
        (0, 1): False,
        (0, None): True,
        (1, 0): False,
        (1, None): False,
        (None, 0): True,
        (None, 1): False,
    },
    ("ccz", "c3sx"): {
        (0, 1, 2): True,
        (0, 1, 3): False,
        (0, 1, None): True,
        (0, 2, 1): True,
        (0, 2, 3): False,
        (0, 2, None): True,
        (0, 3, 1): False,
        (0, 3, 2): False,
        (0, 3, None): False,
        (0, None, 1): True,
        (0, None, 2): True,
        (0, None, 3): False,
        (0, None, None): True,
        (1, 0, 2): True,
        (1, 0, 3): False,
        (1, 0, None): True,
        (1, 2, 0): True,
        (1, 2, 3): False,
        (1, 2, None): True,
        (1, 3, 0): False,
        (1, 3, 2): False,
        (1, 3, None): False,
        (1, None, 0): True,
        (1, None, 2): True,
        (1, None, 3): False,
        (1, None, None): True,
        (2, 0, 1): True,
        (2, 0, 3): False,
        (2, 0, None): True,
        (2, 1, 0): True,
        (2, 1, 3): False,
        (2, 1, None): True,
        (2, 3, 0): False,
        (2, 3, 1): False,
        (2, 3, None): False,
        (2, None, 0): True,
        (2, None, 1): True,
        (2, None, 3): False,
        (2, None, None): True,
        (3, 0, 1): False,
        (3, 0, 2): False,
        (3, 0, None): False,
        (3, 1, 0): False,
        (3, 1, 2): False,
        (3, 1, None): False,
        (3, 2, 0): False,
        (3, 2, 1): False,
        (3, 2, None): False,
        (3, None, 0): False,
        (3, None, 1): False,
        (3, None, 2): False,
        (3, None, None): False,
        (None, 0, 1): True,
        (None, 0, 2): True,
        (None, 0, 3): False,
        (None, 0, None): True,
        (None, 1, 0): True,
        (None, 1, 2): True,
        (None, 1, 3): False,
        (None, 1, None): True,
        (None, 2, 0): True,
        (None, 2, 1): True,
        (None, 2, 3): False,
        (None, 2, None): True,
        (None, 3, 0): False,
        (None, 3, 1): False,
        (None, 3, 2): False,
        (None, 3, None): False,
        (None, None, 0): True,
        (None, None, 1): True,
        (None, None, 2): True,
        (None, None, 3): False,
    },
    ("ccz", "cswap"): {
        (0, 1, 2): True,
        (0, 1, None): False,
        (0, 2, 1): True,
        (0, 2, None): False,
        (0, None, 1): False,
        (0, None, 2): False,
        (0, None, None): True,
        (1, 0, 2): True,
        (1, 0, None): False,
        (1, 2, 0): True,
        (1, 2, None): True,
        (1, None, 0): False,
        (1, None, 2): True,
        (1, None, None): False,
        (2, 0, 1): True,
        (2, 0, None): False,
        (2, 1, 0): True,
        (2, 1, None): True,
        (2, None, 0): False,
        (2, None, 1): True,
        (2, None, None): False,
        (None, 0, 1): False,
        (None, 0, 2): False,
        (None, 0, None): True,
        (None, 1, 0): False,
        (None, 1, 2): True,
        (None, 1, None): False,
        (None, 2, 0): False,
        (None, 2, 1): True,
        (None, 2, None): False,
        (None, None, 0): True,
        (None, None, 1): False,
        (None, None, 2): False,
    },
    ("ccz", "ccz"): True,
    ("ccz", "rccx"): {
        (0, 1, 2): False,
        (0, 1, None): True,
        (0, 2, 1): False,
        (0, 2, None): False,
        (0, None, 1): True,
        (0, None, 2): False,
        (0, None, None): True,
        (1, 0, 2): False,
        (1, 0, None): True,
        (1, 2, 0): False,
        (1, 2, None): False,
        (1, None, 0): True,
        (1, None, 2): False,
        (1, None, None): True,
        (2, 0, 1): False,
        (2, 0, None): False,
        (2, 1, 0): False,
        (2, 1, None): False,
        (2, None, 0): False,
        (2, None, 1): False,
        (2, None, None): False,
        (None, 0, 1): True,
        (None, 0, 2): False,
        (None, 0, None): True,
        (None, 1, 0): True,
        (None, 1, 2): False,
        (None, 1, None): True,
        (None, 2, 0): False,
        (None, 2, 1): False,
        (None, 2, None): False,
        (None, None, 0): True,
        (None, None, 1): True,
        (None, None, 2): False,
    },
    ("ccz", "rcccx"): {
        (0, 1, 2): True,
        (0, 1, 3): False,
        (0, 1, None): True,
        (0, 2, 1): True,
        (0, 2, 3): False,
        (0, 2, None): True,
        (0, 3, 1): False,
        (0, 3, 2): False,
        (0, 3, None): False,
        (0, None, 1): True,
        (0, None, 2): True,
        (0, None, 3): False,
        (0, None, None): True,
        (1, 0, 2): True,
        (1, 0, 3): False,
        (1, 0, None): True,
        (1, 2, 0): True,
        (1, 2, 3): False,
        (1, 2, None): True,
        (1, 3, 0): False,
        (1, 3, 2): False,
        (1, 3, None): False,
        (1, None, 0): True,
        (1, None, 2): True,
        (1, None, 3): False,
        (1, None, None): True,
        (2, 0, 1): True,
        (2, 0, 3): False,
        (2, 0, None): True,
        (2, 1, 0): True,
        (2, 1, 3): False,
        (2, 1, None): True,
        (2, 3, 0): False,
        (2, 3, 1): False,
        (2, 3, None): False,
        (2, None, 0): True,
        (2, None, 1): True,
        (2, None, 3): False,
        (2, None, None): True,
        (3, 0, 1): False,
        (3, 0, 2): False,
        (3, 0, None): False,
        (3, 1, 0): False,
        (3, 1, 2): False,
        (3, 1, None): False,
        (3, 2, 0): False,
        (3, 2, 1): False,
        (3, 2, None): False,
        (3, None, 0): False,
        (3, None, 1): False,
        (3, None, 2): False,
        (3, None, None): False,
        (None, 0, 1): True,
        (None, 0, 2): True,
        (None, 0, 3): False,
        (None, 0, None): True,
        (None, 1, 0): True,
        (None, 1, 2): True,
        (None, 1, 3): False,
        (None, 1, None): True,
        (None, 2, 0): True,
        (None, 2, 1): True,
        (None, 2, 3): False,
        (None, 2, None): True,
        (None, 3, 0): False,
        (None, 3, 1): False,
        (None, 3, 2): False,
        (None, 3, None): False,
        (None, None, 0): True,
        (None, None, 1): True,
        (None, None, 2): True,
        (None, None, 3): False,
    },
    ("h", "id"): True,
    ("h", "sx"): False,
    ("h", "x"): False,
    ("h", "cx"): False,
    ("h", "c3sx"): False,
    ("h", "ccx"): False,
    ("h", "dcx"): False,
    ("h", "ch"): {
        (0,): False,
        (1,): True,
    },
    ("h", "cswap"): False,
    ("h", "csx"): False,
    ("h", "cy"): False,
    ("h", "cz"): False,
    ("h", "ccz"): False,
    ("h", "h"): True,
    ("h", "rccx"): False,
    ("h", "rcccx"): False,
    ("h", "ecr"): False,
    ("h", "s"): False,
    ("h", "sdg"): False,
    ("h", "cs"): False,
    ("h", "csdg"): False,
    ("h", "swap"): False,
    ("h", "iswap"): False,
    ("h", "sxdg"): False,
    ("h", "t"): False,
    ("h", "tdg"): False,
    ("h", "y"): False,
    ("h", "z"): False,
    ("h", "rxx"): False,
    ("h", "ryy"): False,
    ("h", "rzz"): False,
    ("h", "rzx"): False,
    ("rccx", "c3sx"): {
        (0, 1, 2): False,
        (0, 1, 3): False,
        (0, 1, None): True,
        (0, 2, 1): False,
        (0, 2, 3): False,
        (0, 2, None): True,
        (0, 3, 1): False,
        (0, 3, 2): False,
        (0, 3, None): False,
        (0, None, 1): False,
        (0, None, 2): False,
        (0, None, 3): False,
        (0, None, None): True,
        (1, 0, 2): False,
        (1, 0, 3): False,
        (1, 0, None): True,
        (1, 2, 0): False,
        (1, 2, 3): False,
        (1, 2, None): True,
        (1, 3, 0): False,
        (1, 3, 2): False,
        (1, 3, None): False,
        (1, None, 0): False,
        (1, None, 2): False,
        (1, None, 3): False,
        (1, None, None): True,
        (2, 0, 1): False,
        (2, 0, 3): False,
        (2, 0, None): True,
        (2, 1, 0): False,
        (2, 1, 3): False,
        (2, 1, None): True,
        (2, 3, 0): False,
        (2, 3, 1): False,
        (2, 3, None): False,
        (2, None, 0): False,
        (2, None, 1): False,
        (2, None, 3): False,
        (2, None, None): True,
        (3, 0, 1): False,
        (3, 0, 2): False,
        (3, 0, None): False,
        (3, 1, 0): False,
        (3, 1, 2): False,
        (3, 1, None): False,
        (3, 2, 0): False,
        (3, 2, 1): False,
        (3, 2, None): False,
        (3, None, 0): False,
        (3, None, 1): False,
        (3, None, 2): False,
        (3, None, None): False,
        (None, 0, 1): False,
        (None, 0, 2): False,
        (None, 0, 3): False,
        (None, 0, None): True,
        (None, 1, 0): False,
        (None, 1, 2): False,
        (None, 1, 3): False,
        (None, 1, None): True,
        (None, 2, 0): False,
        (None, 2, 1): False,
        (None, 2, 3): False,
        (None, 2, None): True,
        (None, 3, 0): False,
        (None, 3, 1): False,
        (None, 3, 2): False,
        (None, 3, None): False,
        (None, None, 0): False,
        (None, None, 1): False,
        (None, None, 2): False,
        (None, None, 3): False,
    },
    ("rccx", "cswap"): {
        (0, 1, 2): False,
        (0, 1, None): False,
        (0, 2, 1): False,
        (0, 2, None): False,
        (0, None, 1): False,
        (0, None, 2): False,
        (0, None, None): True,
        (1, 0, 2): False,
        (1, 0, None): False,
        (1, 2, 0): False,
        (1, 2, None): False,
        (1, None, 0): False,
        (1, None, 2): False,
        (1, None, None): False,
        (2, 0, 1): False,
        (2, 0, None): False,
        (2, 1, 0): False,
        (2, 1, None): False,
        (2, None, 0): False,
        (2, None, 1): False,
        (2, None, None): False,
        (None, 0, 1): False,
        (None, 0, 2): False,
        (None, 0, None): True,
        (None, 1, 0): False,
        (None, 1, 2): False,
        (None, 1, None): False,
        (None, 2, 0): False,
        (None, 2, 1): False,
        (None, 2, None): False,
        (None, None, 0): False,
        (None, None, 1): False,
        (None, None, 2): False,
    },
    ("rccx", "rccx"): {
        (0, 1, 2): True,
        (0, 1, None): True,
        (0, 2, 1): False,
        (0, 2, None): False,
        (0, None, 1): False,
        (0, None, 2): False,
        (0, None, None): True,
        (1, 0, 2): True,
        (1, 0, None): True,
        (1, 2, 0): False,
        (1, 2, None): False,
        (1, None, 0): False,
        (1, None, 2): False,
        (1, None, None): True,
        (2, 0, 1): False,
        (2, 0, None): False,
        (2, 1, 0): False,
        (2, 1, None): False,
        (2, None, 0): False,
        (2, None, 1): False,
        (2, None, None): False,
        (None, 0, 1): False,
        (None, 0, 2): False,
        (None, 0, None): True,
        (None, 1, 0): False,
        (None, 1, 2): True,
        (None, 1, None): True,
        (None, 2, 0): False,
        (None, 2, 1): False,
        (None, 2, None): False,
        (None, None, 0): False,
        (None, None, 1): False,
        (None, None, 2): False,
    },
    ("rccx", "rcccx"): {
        (0, 1, 2): False,
        (0, 1, 3): False,
        (0, 1, None): True,
        (0, 2, 1): False,
        (0, 2, 3): True,
        (0, 2, None): True,
        (0, 3, 1): False,
        (0, 3, 2): False,
        (0, 3, None): False,
        (0, None, 1): False,
        (0, None, 2): False,
        (0, None, 3): False,
        (0, None, None): True,
        (1, 0, 2): False,
        (1, 0, 3): False,
        (1, 0, None): True,
        (1, 2, 0): False,
        (1, 2, 3): True,
        (1, 2, None): True,
        (1, 3, 0): False,
        (1, 3, 2): False,
        (1, 3, None): False,
        (1, None, 0): False,
        (1, None, 2): False,
        (1, None, 3): False,
        (1, None, None): True,
        (2, 0, 1): False,
        (2, 0, 3): True,
        (2, 0, None): True,
        (2, 1, 0): False,
        (2, 1, 3): True,
        (2, 1, None): True,
        (2, 3, 0): False,
        (2, 3, 1): False,
        (2, 3, None): False,
        (2, None, 0): False,
        (2, None, 1): False,
        (2, None, 3): False,
        (2, None, None): True,
        (3, 0, 1): False,
        (3, 0, 2): False,
        (3, 0, None): False,
        (3, 1, 0): False,
        (3, 1, 2): False,
        (3, 1, None): False,
        (3, 2, 0): False,
        (3, 2, 1): False,
        (3, 2, None): False,
        (3, None, 0): False,
        (3, None, 1): False,
        (3, None, 2): False,
        (3, None, None): False,
        (None, 0, 1): False,
        (None, 0, 2): False,
        (None, 0, 3): False,
        (None, 0, None): True,
        (None, 1, 0): False,
        (None, 1, 2): False,
        (None, 1, 3): False,
        (None, 1, None): True,
        (None, 2, 0): False,
        (None, 2, 1): False,
        (None, 2, 3): True,
        (None, 2, None): True,
        (None, 3, 0): False,
        (None, 3, 1): False,
        (None, 3, 2): False,
        (None, 3, None): False,
        (None, None, 0): False,
        (None, None, 1): False,
        (None, None, 2): False,
        (None, None, 3): False,
    },
    ("rcccx", "rcccx"): {
        (0, 1, 2, 3): True,
        (0, 1, 2, None): True,
        (0, 1, 3, 2): False,
        (0, 1, 3, None): False,
        (0, 1, None, 2): False,
        (0, 1, None, 3): False,
        (0, 1, None, None): True,
        (0, 2, 1, 3): True,
        (0, 2, 1, None): True,
        (0, 2, 3, 1): False,
        (0, 2, 3, None): False,
        (0, 2, None, 1): False,
        (0, 2, None, 3): False,
        (0, 2, None, None): True,
        (0, 3, 1, 2): False,
        (0, 3, 1, None): False,
        (0, 3, 2, 1): False,
        (0, 3, 2, None): False,
        (0, 3, None, 1): False,
        (0, 3, None, 2): False,
        (0, 3, None, None): False,
        (0, None, 1, 2): False,
        (0, None, 1, 3): False,
        (0, None, 1, None): True,
        (0, None, 2, 1): False,
        (0, None, 2, 3): True,
        (0, None, 2, None): True,
        (0, None, 3, 1): False,
        (0, None, 3, 2): False,
        (0, None, 3, None): False,
        (0, None, None, 1): False,
        (0, None, None, 2): False,
        (0, None, None, 3): False,
        (0, None, None, None): True,
        (1, 0, 2, 3): True,
        (1, 0, 2, None): True,
        (1, 0, 3, 2): False,
        (1, 0, 3, None): False,
        (1, 0, None, 2): False,
        (1, 0, None, 3): False,
        (1, 0, None, None): True,
        (1, 2, 0, 3): True,
        (1, 2, 0, None): True,
        (1, 2, 3, 0): False,
        (1, 2, 3, None): False,
        (1, 2, None, 0): False,
        (1, 2, None, 3): False,
        (1, 2, None, None): True,
        (1, 3, 0, 2): False,
        (1, 3, 0, None): False,
        (1, 3, 2, 0): False,
        (1, 3, 2, None): False,
        (1, 3, None, 0): False,
        (1, 3, None, 2): False,
        (1, 3, None, None): False,
        (1, None, 0, 2): False,
        (1, None, 0, 3): False,
        (1, None, 0, None): True,
        (1, None, 2, 0): False,
        (1, None, 2, 3): True,
        (1, None, 2, None): True,
        (1, None, 3, 0): False,
        (1, None, 3, 2): False,
        (1, None, 3, None): False,
        (1, None, None, 0): False,
        (1, None, None, 2): False,
        (1, None, None, 3): False,
        (1, None, None, None): True,
        (2, 0, 1, 3): True,
        (2, 0, 1, None): True,
        (2, 0, 3, 1): False,
        (2, 0, 3, None): False,
        (2, 0, None, 1): False,
        (2, 0, None, 3): False,
        (2, 0, None, None): True,
        (2, 1, 0, 3): True,
        (2, 1, 0, None): True,
        (2, 1, 3, 0): False,
        (2, 1, 3, None): False,
        (2, 1, None, 0): False,
        (2, 1, None, 3): False,
        (2, 1, None, None): True,
        (2, 3, 0, 1): False,
        (2, 3, 0, None): False,
        (2, 3, 1, 0): False,
        (2, 3, 1, None): False,
        (2, 3, None, 0): False,
        (2, 3, None, 1): False,
        (2, 3, None, None): False,
        (2, None, 0, 1): False,
        (2, None, 0, 3): True,
        (2, None, 0, None): True,
        (2, None, 1, 0): False,
        (2, None, 1, 3): True,
        (2, None, 1, None): True,
        (2, None, 3, 0): False,
        (2, None, 3, 1): False,
        (2, None, 3, None): False,
        (2, None, None, 0): False,
        (2, None, None, 1): False,
        (2, None, None, 3): False,
        (2, None, None, None): True,
        (3, 0, 1, 2): False,
        (3, 0, 1, None): False,
        (3, 0, 2, 1): False,
        (3, 0, 2, None): False,
        (3, 0, None, 1): False,
        (3, 0, None, 2): False,
        (3, 0, None, None): False,
        (3, 1, 0, 2): False,
        (3, 1, 0, None): False,
        (3, 1, 2, 0): False,
        (3, 1, 2, None): False,
        (3, 1, None, 0): False,
        (3, 1, None, 2): False,
        (3, 1, None, None): False,
        (3, 2, 0, 1): False,
        (3, 2, 0, None): False,
        (3, 2, 1, 0): False,
        (3, 2, 1, None): False,
        (3, 2, None, 0): False,
        (3, 2, None, 1): False,
        (3, 2, None, None): False,
        (3, None, 0, 1): False,
        (3, None, 0, 2): False,
        (3, None, 0, None): False,
        (3, None, 1, 0): False,
        (3, None, 1, 2): False,
        (3, None, 1, None): False,
        (3, None, 2, 0): False,
        (3, None, 2, 1): False,
        (3, None, 2, None): False,
        (3, None, None, 0): False,
        (3, None, None, 1): False,
        (3, None, None, 2): False,
        (3, None, None, None): False,
        (None, 0, 1, 2): False,
        (None, 0, 1, 3): False,
        (None, 0, 1, None): True,
        (None, 0, 2, 1): False,
        (None, 0, 2, 3): True,
        (None, 0, 2, None): True,
        (None, 0, 3, 1): False,
        (None, 0, 3, 2): False,
        (None, 0, 3, None): False,
        (None, 0, None, 1): False,
        (None, 0, None, 2): False,
        (None, 0, None, 3): False,
        (None, 0, None, None): True,
        (None, 1, 0, 2): False,
        (None, 1, 0, 3): False,
        (None, 1, 0, None): True,
        (None, 1, 2, 0): False,
        (None, 1, 2, 3): True,
        (None, 1, 2, None): True,
        (None, 1, 3, 0): False,
        (None, 1, 3, 2): False,
        (None, 1, 3, None): False,
        (None, 1, None, 0): False,
        (None, 1, None, 2): False,
        (None, 1, None, 3): False,
        (None, 1, None, None): True,
        (None, 2, 0, 1): False,
        (None, 2, 0, 3): True,
        (None, 2, 0, None): True,
        (None, 2, 1, 0): False,
        (None, 2, 1, 3): True,
        (None, 2, 1, None): True,
        (None, 2, 3, 0): False,
        (None, 2, 3, 1): False,
        (None, 2, 3, None): False,
        (None, 2, None, 0): False,
        (None, 2, None, 1): False,
        (None, 2, None, 3): False,
        (None, 2, None, None): True,
        (None, 3, 0, 1): False,
        (None, 3, 0, 2): False,
        (None, 3, 0, None): False,
        (None, 3, 1, 0): False,
        (None, 3, 1, 2): False,
        (None, 3, 1, None): False,
        (None, 3, 2, 0): False,
        (None, 3, 2, 1): False,
        (None, 3, 2, None): False,
        (None, 3, None, 0): False,
        (None, 3, None, 1): False,
        (None, 3, None, 2): False,
        (None, 3, None, None): False,
        (None, None, 0, 1): False,
        (None, None, 0, 2): False,
        (None, None, 0, 3): False,
        (None, None, 0, None): True,
        (None, None, 1, 0): False,
        (None, None, 1, 2): False,
        (None, None, 1, 3): False,
        (None, None, 1, None): True,
        (None, None, 2, 0): False,
        (None, None, 2, 1): False,
        (None, None, 2, 3): True,
        (None, None, 2, None): True,
        (None, None, 3, 0): False,
        (None, None, 3, 1): False,
        (None, None, 3, 2): False,
        (None, None, 3, None): False,
        (None, None, None, 0): False,
        (None, None, None, 1): False,
        (None, None, None, 2): False,
        (None, None, None, 3): False,
    },
    ("ecr", "c3sx"): {
        (0, 1): False,
        (0, 2): False,
        (0, 3): False,
        (0, None): False,
        (1, 0): False,
        (1, 2): False,
        (1, 3): False,
        (1, None): False,
        (2, 0): False,
        (2, 1): False,
        (2, 3): False,
        (2, None): False,
        (3, 0): False,
        (3, 1): False,
        (3, 2): False,
        (3, None): False,
        (None, 0): False,
        (None, 1): False,
        (None, 2): False,
        (None, 3): True,
    },
    ("ecr", "ccx"): {
        (0, 1): False,
        (0, 2): False,
        (0, None): False,
        (1, 0): False,
        (1, 2): False,
        (1, None): False,
        (2, 0): False,
        (2, 1): False,
        (2, None): False,
        (None, 0): False,
        (None, 1): False,
        (None, 2): True,
    },
    ("ecr", "cswap"): False,
    ("ecr", "ccz"): False,
    ("ecr", "rccx"): False,
    ("ecr", "rcccx"): False,
    ("ecr", "ecr"): {
        (0, 1): True,
        (0, None): False,
        (1, 0): True,
        (1, None): False,
        (None, 0): False,
        (None, 1): True,
    },
    ("ecr", "csdg"): False,
    ("ecr", "swap"): False,
    ("ecr", "iswap"): False,
    ("ecr", "rxx"): {
        (0, 1): False,
        (0, None): False,
        (1, 0): False,
        (1, None): False,
        (None, 0): True,
        (None, 1): True,
    },
    ("ecr", "ryy"): False,
    ("ecr", "rzz"): False,
    ("ecr", "rzx"): {
        (0, 1): False,
        (0, None): False,
        (1, 0): True,
        (1, None): False,
        (None, 0): False,
        (None, 1): True,
    },
    ("s", "id"): True,
    ("s", "sx"): False,
    ("s", "x"): False,
    ("s", "cx"): {
        (0,): True,
        (1,): False,
    },
    ("s", "c3sx"): {
        (0,): True,
        (1,): True,
        (2,): True,
        (3,): False,
    },
    ("s", "ccx"): {
        (0,): True,
        (1,): True,
        (2,): False,
    },
    ("s", "dcx"): False,
    ("s", "ch"): {
        (0,): True,
        (1,): False,
    },
    ("s", "cswap"): {
        (0,): True,
        (1,): False,
        (2,): False,
    },
    ("s", "csx"): {
        (0,): True,
        (1,): False,
    },
    ("s", "cy"): {
        (0,): True,
        (1,): False,
    },
    ("s", "cz"): True,
    ("s", "ccz"): True,
    ("s", "rccx"): {
        (0,): True,
        (1,): True,
        (2,): False,
    },
    ("s", "rcccx"): {
        (0,): True,
        (1,): True,
        (2,): True,
        (3,): False,
    },
    ("s", "ecr"): False,
    ("s", "s"): True,
    ("s", "sdg"): True,
    ("s", "cs"): True,
    ("s", "csdg"): True,
    ("s", "swap"): False,
    ("s", "iswap"): False,
    ("s", "sxdg"): False,
    ("s", "t"): True,
    ("s", "tdg"): True,
    ("s", "y"): False,
    ("s", "z"): True,
    ("s", "rxx"): False,
    ("s", "ryy"): False,
    ("s", "rzz"): True,
    ("s", "rzx"): {
        (0,): True,
        (1,): False,
    },
    ("sdg", "cx"): {
        (0,): True,
        (1,): False,
    },
    ("sdg", "c3sx"): {
        (0,): True,
        (1,): True,
        (2,): True,
        (3,): False,
    },
    ("sdg", "ccx"): {
        (0,): True,
        (1,): True,
        (2,): False,
    },
    ("sdg", "dcx"): False,
    ("sdg", "ch"): {
        (0,): True,
        (1,): False,
    },
    ("sdg", "cswap"): {
        (0,): True,
        (1,): False,
        (2,): False,
    },
    ("sdg", "csx"): {
        (0,): True,
        (1,): False,
    },
    ("sdg", "cy"): {
        (0,): True,
        (1,): False,
    },
    ("sdg", "cz"): True,
    ("sdg", "ccz"): True,
    ("sdg", "rccx"): {
        (0,): True,
        (1,): True,
        (2,): False,
    },
    ("sdg", "rcccx"): {
        (0,): True,
        (1,): True,
        (2,): True,
        (3,): False,
    },
    ("sdg", "ecr"): False,
    ("sdg", "sdg"): True,
    ("sdg", "cs"): True,
    ("sdg", "csdg"): True,
    ("sdg", "swap"): False,
    ("sdg", "iswap"): False,
    ("sdg", "sxdg"): False,
    ("sdg", "tdg"): True,
    ("sdg", "rxx"): False,
    ("sdg", "ryy"): False,
    ("sdg", "rzz"): True,
    ("sdg", "rzx"): {
        (0,): True,
        (1,): False,
    },
    ("cs", "cx"): {
        (0, 1): False,
        (0, None): True,
        (1, 0): False,
        (1, None): False,
        (None, 0): True,
        (None, 1): False,
    },
    ("cs", "c3sx"): {
        (0, 1): True,
        (0, 2): True,
        (0, 3): False,
        (0, None): True,
        (1, 0): True,
        (1, 2): True,
        (1, 3): False,
        (1, None): True,
        (2, 0): True,
        (2, 1): True,
        (2, 3): False,
        (2, None): True,
        (3, 0): False,
        (3, 1): False,
        (3, 2): False,
        (3, None): False,
        (None, 0): True,
        (None, 1): True,
        (None, 2): True,
        (None, 3): False,
    },
    ("cs", "ccx"): {
        (0, 1): True,
        (0, 2): False,
        (0, None): True,
        (1, 0): True,
        (1, 2): False,
        (1, None): True,
        (2, 0): False,
        (2, 1): False,
        (2, None): False,
        (None, 0): True,
        (None, 1): True,
        (None, 2): False,
    },
    ("cs", "dcx"): False,
    ("cs", "cswap"): {
        (0, 1): False,
        (0, 2): False,
        (0, None): True,
        (1, 0): False,
        (1, 2): True,
        (1, None): False,
        (2, 0): False,
        (2, 1): True,
        (2, None): False,
        (None, 0): True,
        (None, 1): False,
        (None, 2): False,
    },
    ("cs", "csx"): {
        (0, 1): False,
        (0, None): True,
        (1, 0): False,
        (1, None): False,
        (None, 0): True,
        (None, 1): False,
    },
    ("cs", "cy"): {
        (0, 1): False,
        (0, None): True,
        (1, 0): False,
        (1, None): False,
        (None, 0): True,
        (None, 1): False,
    },
    ("cs", "cz"): True,
    ("cs", "ccz"): True,
    ("cs", "rccx"): {
        (0, 1): True,
        (0, 2): False,
        (0, None): True,
        (1, 0): True,
        (1, 2): False,
        (1, None): True,
        (2, 0): False,
        (2, 1): False,
        (2, None): False,
        (None, 0): True,
        (None, 1): True,
        (None, 2): False,
    },
    ("cs", "rcccx"): {
        (0, 1): True,
        (0, 2): True,
        (0, 3): False,
        (0, None): True,
        (1, 0): True,
        (1, 2): True,
        (1, 3): False,
        (1, None): True,
        (2, 0): True,
        (2, 1): True,
        (2, 3): False,
        (2, None): True,
        (3, 0): False,
        (3, 1): False,
        (3, 2): False,
        (3, None): False,
        (None, 0): True,
        (None, 1): True,
        (None, 2): True,
        (None, 3): False,
    },
    ("cs", "ecr"): False,
    ("cs", "cs"): True,
    ("cs", "csdg"): True,
    ("cs", "swap"): {
        (0, 1): True,
        (0, None): False,
        (1, 0): True,
        (1, None): False,
        (None, 0): False,
        (None, 1): False,
    },
    ("cs", "iswap"): {
        (0, 1): True,
        (0, None): False,
        (1, 0): True,
        (1, None): False,
        (None, 0): False,
        (None, 1): False,
    },
    ("cs", "rxx"): False,
    ("cs", "ryy"): False,
    ("cs", "rzz"): True,
    ("cs", "rzx"): {
        (0, 1): False,
        (0, None): True,
        (1, 0): False,
        (1, None): False,
        (None, 0): True,
        (None, 1): False,
    },
    ("csdg", "c3sx"): {
        (0, 1): True,
        (0, 2): True,
        (0, 3): False,
        (0, None): True,
        (1, 0): True,
        (1, 2): True,
        (1, 3): False,
        (1, None): True,
        (2, 0): True,
        (2, 1): True,
        (2, 3): False,
        (2, None): True,
        (3, 0): False,
        (3, 1): False,
        (3, 2): False,
        (3, None): False,
        (None, 0): True,
        (None, 1): True,
        (None, 2): True,
        (None, 3): False,
    },
    ("csdg", "ccx"): {
        (0, 1): True,
        (0, 2): False,
        (0, None): True,
        (1, 0): True,
        (1, 2): False,
        (1, None): True,
        (2, 0): False,
        (2, 1): False,
        (2, None): False,
        (None, 0): True,
        (None, 1): True,
        (None, 2): False,
    },
    ("csdg", "cswap"): {
        (0, 1): False,
        (0, 2): False,
        (0, None): True,
        (1, 0): False,
        (1, 2): True,
        (1, None): False,
        (2, 0): False,
        (2, 1): True,
        (2, None): False,
        (None, 0): True,
        (None, 1): False,
        (None, 2): False,
    },
    ("csdg", "ccz"): True,
    ("csdg", "rccx"): {
        (0, 1): True,
        (0, 2): False,
        (0, None): True,
        (1, 0): True,
        (1, 2): False,
        (1, None): True,
        (2, 0): False,
        (2, 1): False,
        (2, None): False,
        (None, 0): True,
        (None, 1): True,
        (None, 2): False,
    },
    ("csdg", "rcccx"): {
        (0, 1): True,
        (0, 2): True,
        (0, 3): False,
        (0, None): True,
        (1, 0): True,
        (1, 2): True,
        (1, 3): False,
        (1, None): True,
        (2, 0): True,
        (2, 1): True,
        (2, 3): False,
        (2, None): True,
        (3, 0): False,
        (3, 1): False,
        (3, 2): False,
        (3, None): False,
        (None, 0): True,
        (None, 1): True,
        (None, 2): True,
        (None, 3): False,
    },
    ("csdg", "csdg"): True,
    ("csdg", "swap"): {
        (0, 1): True,
        (0, None): False,
        (1, 0): True,
        (1, None): False,
        (None, 0): False,
        (None, 1): False,
    },
    ("csdg", "iswap"): {
        (0, 1): True,
        (0, None): False,
        (1, 0): True,
        (1, None): False,
        (None, 0): False,
        (None, 1): False,
    },
    ("swap", "c3sx"): {
        (0, 1): True,
        (0, 2): True,
        (0, 3): False,
        (0, None): False,
        (1, 0): True,
        (1, 2): True,
        (1, 3): False,
        (1, None): False,
        (2, 0): True,
        (2, 1): True,
        (2, 3): False,
        (2, None): False,
        (3, 0): False,
        (3, 1): False,
        (3, 2): False,
        (3, None): False,
        (None, 0): False,
        (None, 1): False,
        (None, 2): False,
        (None, 3): False,
    },
    ("swap", "ccx"): {
        (0, 1): True,
        (0, 2): False,
        (0, None): False,
        (1, 0): True,
        (1, 2): False,
        (1, None): False,
        (2, 0): False,
        (2, 1): False,
        (2, None): False,
        (None, 0): False,
        (None, 1): False,
        (None, 2): False,
    },
    ("swap", "cswap"): {
        (0, 1): False,
        (0, 2): False,
        (0, None): False,
        (1, 0): False,
        (1, 2): True,
        (1, None): False,
        (2, 0): False,
        (2, 1): True,
        (2, None): False,
        (None, 0): False,
        (None, 1): False,
        (None, 2): False,
    },
    ("swap", "ccz"): {
        (0, 1): True,
        (0, 2): True,
        (0, None): False,
        (1, 0): True,
        (1, 2): True,
        (1, None): False,
        (2, 0): True,
        (2, 1): True,
        (2, None): False,
        (None, 0): False,
        (None, 1): False,
        (None, 2): False,
    },
    ("swap", "rccx"): False,
    ("swap", "rcccx"): {
        (0, 1): True,
        (0, 2): False,
        (0, 3): False,
        (0, None): False,
        (1, 0): True,
        (1, 2): False,
        (1, 3): False,
        (1, None): False,
        (2, 0): False,
        (2, 1): False,
        (2, 3): False,
        (2, None): False,
        (3, 0): False,
        (3, 1): False,
        (3, 2): False,
        (3, None): False,
        (None, 0): False,
        (None, 1): False,
        (None, 2): False,
        (None, 3): False,
    },
    ("swap", "swap"): {
        (0, 1): True,
        (0, None): False,
        (1, 0): True,
        (1, None): False,
        (None, 0): False,
        (None, 1): False,
    },
    ("swap", "iswap"): {
        (0, 1): True,
        (0, None): False,
        (1, 0): True,
        (1, None): False,
        (None, 0): False,
        (None, 1): False,
    },
    ("iswap", "c3sx"): {
        (0, 1): True,
        (0, 2): True,
        (0, 3): False,
        (0, None): False,
        (1, 0): True,
        (1, 2): True,
        (1, 3): False,
        (1, None): False,
        (2, 0): True,
        (2, 1): True,
        (2, 3): False,
        (2, None): False,
        (3, 0): False,
        (3, 1): False,
        (3, 2): False,
        (3, None): False,
        (None, 0): False,
        (None, 1): False,
        (None, 2): False,
        (None, 3): False,
    },
    ("iswap", "ccx"): {
        (0, 1): True,
        (0, 2): False,
        (0, None): False,
        (1, 0): True,
        (1, 2): False,
        (1, None): False,
        (2, 0): False,
        (2, 1): False,
        (2, None): False,
        (None, 0): False,
        (None, 1): False,
        (None, 2): False,
    },
    ("iswap", "cswap"): {
        (0, 1): False,
        (0, 2): False,
        (0, None): False,
        (1, 0): False,
        (1, 2): True,
        (1, None): False,
        (2, 0): False,
        (2, 1): True,
        (2, None): False,
        (None, 0): False,
        (None, 1): False,
        (None, 2): False,
    },
    ("iswap", "ccz"): {
        (0, 1): True,
        (0, 2): True,
        (0, None): False,
        (1, 0): True,
        (1, 2): True,
        (1, None): False,
        (2, 0): True,
        (2, 1): True,
        (2, None): False,
        (None, 0): False,
        (None, 1): False,
        (None, 2): False,
    },
    ("iswap", "rccx"): False,
    ("iswap", "rcccx"): {
        (0, 1): True,
        (0, 2): False,
        (0, 3): False,
        (0, None): False,
        (1, 0): True,
        (1, 2): False,
        (1, 3): False,
        (1, None): False,
        (2, 0): False,
        (2, 1): False,
        (2, 3): False,
        (2, None): False,
        (3, 0): False,
        (3, 1): False,
        (3, 2): False,
        (3, None): False,
        (None, 0): False,
        (None, 1): False,
        (None, 2): False,
        (None, 3): False,
    },
    ("iswap", "iswap"): {
        (0, 1): True,
        (0, None): False,
        (1, 0): True,
        (1, None): False,
        (None, 0): False,
        (None, 1): False,
    },
    ("sxdg", "cx"): {
        (0,): False,
        (1,): True,
    },
    ("sxdg", "c3sx"): {
        (0,): False,
        (1,): False,
        (2,): False,
        (3,): True,
    },
    ("sxdg", "ccx"): {
        (0,): False,
        (1,): False,
        (2,): True,
    },
    ("sxdg", "dcx"): False,
    ("sxdg", "ch"): False,
    ("sxdg", "cswap"): False,
    ("sxdg", "csx"): {
        (0,): False,
        (1,): True,
    },
    ("sxdg", "cy"): False,
    ("sxdg", "cz"): False,
    ("sxdg", "ccz"): False,
    ("sxdg", "rccx"): False,
    ("sxdg", "rcccx"): False,
    ("sxdg", "ecr"): {
        (0,): False,
        (1,): True,
    },
    ("sxdg", "cs"): False,
    ("sxdg", "csdg"): False,
    ("sxdg", "swap"): False,
    ("sxdg", "iswap"): False,
    ("sxdg", "sxdg"): True,
    ("sxdg", "rxx"): True,
    ("sxdg", "ryy"): False,
    ("sxdg", "rzz"): False,
    ("sxdg", "rzx"): {
        (0,): False,
        (1,): True,
    },
    ("t", "id"): True,
    ("t", "sx"): False,
    ("t", "x"): False,
    ("t", "cx"): {
        (0,): True,
        (1,): False,
    },
    ("t", "c3sx"): {
        (0,): True,
        (1,): True,
        (2,): True,
        (3,): False,
    },
    ("t", "ccx"): {
        (0,): True,
        (1,): True,
        (2,): False,
    },
    ("t", "dcx"): False,
    ("t", "ch"): {
        (0,): True,
        (1,): False,
    },
    ("t", "cswap"): {
        (0,): True,
        (1,): False,
        (2,): False,
    },
    ("t", "csx"): {
        (0,): True,
        (1,): False,
    },
    ("t", "cy"): {
        (0,): True,
        (1,): False,
    },
    ("t", "cz"): True,
    ("t", "ccz"): True,
    ("t", "rccx"): {
        (0,): True,
        (1,): True,
        (2,): False,
    },
    ("t", "rcccx"): {
        (0,): True,
        (1,): True,
        (2,): True,
        (3,): False,
    },
    ("t", "ecr"): False,
    ("t", "sdg"): True,
    ("t", "cs"): True,
    ("t", "csdg"): True,
    ("t", "swap"): False,
    ("t", "iswap"): False,
    ("t", "sxdg"): False,
    ("t", "t"): True,
    ("t", "tdg"): True,
    ("t", "y"): False,
    ("t", "z"): True,
    ("t", "rxx"): False,
    ("t", "ryy"): False,
    ("t", "rzz"): True,
    ("t", "rzx"): {
        (0,): True,
        (1,): False,
    },
    ("tdg", "cx"): {
        (0,): True,
        (1,): False,
    },
    ("tdg", "c3sx"): {
        (0,): True,
        (1,): True,
        (2,): True,
        (3,): False,
    },
    ("tdg", "ccx"): {
        (0,): True,
        (1,): True,
        (2,): False,
    },
    ("tdg", "dcx"): False,
    ("tdg", "ch"): {
        (0,): True,
        (1,): False,
    },
    ("tdg", "cswap"): {
        (0,): True,
        (1,): False,
        (2,): False,
    },
    ("tdg", "csx"): {
        (0,): True,
        (1,): False,
    },
    ("tdg", "cy"): {
        (0,): True,
        (1,): False,
    },
    ("tdg", "cz"): True,
    ("tdg", "ccz"): True,
    ("tdg", "rccx"): {
        (0,): True,
        (1,): True,
        (2,): False,
    },
    ("tdg", "rcccx"): {
        (0,): True,
        (1,): True,
        (2,): True,
        (3,): False,
    },
    ("tdg", "ecr"): False,
    ("tdg", "cs"): True,
    ("tdg", "csdg"): True,
    ("tdg", "swap"): False,
    ("tdg", "iswap"): False,
    ("tdg", "sxdg"): False,
    ("tdg", "tdg"): True,
    ("tdg", "rxx"): False,
    ("tdg", "ryy"): False,
    ("tdg", "rzz"): True,
    ("tdg", "rzx"): {
        (0,): True,
        (1,): False,
    },
    ("y", "id"): True,
    ("y", "sx"): False,
    ("y", "cx"): False,
    ("y", "c3sx"): False,
    ("y", "ccx"): False,
    ("y", "dcx"): False,
    ("y", "ch"): False,
    ("y", "cswap"): False,
    ("y", "csx"): False,
    ("y", "cy"): {
        (0,): False,
        (1,): True,
    },
    ("y", "cz"): False,
    ("y", "ccz"): False,
    ("y", "rccx"): False,
    ("y", "rcccx"): False,
    ("y", "ecr"): False,
    ("y", "sdg"): False,
    ("y", "cs"): False,
    ("y", "csdg"): False,
    ("y", "swap"): False,
    ("y", "iswap"): False,
    ("y", "sxdg"): False,
    ("y", "tdg"): False,
    ("y", "y"): True,
    ("y", "z"): False,
    ("y", "rxx"): False,
    ("y", "ryy"): True,
    ("y", "rzz"): False,
    ("y", "rzx"): False,
    ("z", "id"): True,
    ("z", "sx"): False,
    ("z", "cx"): {
        (0,): True,
        (1,): False,
    },
    ("z", "c3sx"): {
        (0,): True,
        (1,): True,
        (2,): True,
        (3,): False,
    },
    ("z", "ccx"): {
        (0,): True,
        (1,): True,
        (2,): False,
    },
    ("z", "dcx"): False,
    ("z", "ch"): {
        (0,): True,
        (1,): False,
    },
    ("z", "cswap"): {
        (0,): True,
        (1,): False,
        (2,): False,
    },
    ("z", "csx"): {
        (0,): True,
        (1,): False,
    },
    ("z", "cy"): {
        (0,): True,
        (1,): False,
    },
    ("z", "cz"): True,
    ("z", "ccz"): True,
    ("z", "rccx"): {
        (0,): True,
        (1,): True,
        (2,): False,
    },
    ("z", "rcccx"): {
        (0,): True,
        (1,): True,
        (2,): True,
        (3,): False,
    },
    ("z", "ecr"): False,
    ("z", "sdg"): True,
    ("z", "cs"): True,
    ("z", "csdg"): True,
    ("z", "swap"): False,
    ("z", "iswap"): False,
    ("z", "sxdg"): False,
    ("z", "tdg"): True,
    ("z", "z"): True,
    ("z", "rxx"): False,
    ("z", "ryy"): False,
    ("z", "rzz"): True,
    ("z", "rzx"): {
        (0,): True,
        (1,): False,
    },
    ("rxx", "c3sx"): {
        (0, 1): False,
        (0, 2): False,
        (0, 3): False,
        (0, None): False,
        (1, 0): False,
        (1, 2): False,
        (1, 3): False,
        (1, None): False,
        (2, 0): False,
        (2, 1): False,
        (2, 3): False,
        (2, None): False,
        (3, 0): False,
        (3, 1): False,
        (3, 2): False,
        (3, None): True,
        (None, 0): False,
        (None, 1): False,
        (None, 2): False,
        (None, 3): True,
    },
    ("rxx", "ccx"): {
        (0, 1): False,
        (0, 2): False,
        (0, None): False,
        (1, 0): False,
        (1, 2): False,
        (1, None): False,
        (2, 0): False,
        (2, 1): False,
        (2, None): True,
        (None, 0): False,
        (None, 1): False,
        (None, 2): True,
    },
    ("rxx", "cswap"): {
        (0, 1): False,
        (0, 2): False,
        (0, None): False,
        (1, 0): False,
        (1, 2): True,
        (1, None): False,
        (2, 0): False,
        (2, 1): True,
        (2, None): False,
        (None, 0): False,
        (None, 1): False,
        (None, 2): False,
    },
    ("rxx", "ccz"): False,
    ("rxx", "rccx"): False,
    ("rxx", "rcccx"): False,
    ("rxx", "csdg"): False,
    ("rxx", "swap"): {
        (0, 1): True,
        (0, None): False,
        (1, 0): True,
        (1, None): False,
        (None, 0): False,
        (None, 1): False,
    },
    ("rxx", "iswap"): {
        (0, 1): True,
        (0, None): False,
        (1, 0): True,
        (1, None): False,
        (None, 0): False,
        (None, 1): False,
    },
    ("rxx", "rxx"): True,
    ("rxx", "ryy"): {
        (0, 1): True,
        (0, None): False,
        (1, 0): True,
        (1, None): False,
        (None, 0): False,
        (None, 1): False,
    },
    ("rxx", "rzz"): {
        (0, 1): True,
        (0, None): False,
        (1, 0): True,
        (1, None): False,
        (None, 0): False,
        (None, 1): False,
    },
    ("rxx", "rzx"): {
        (0, 1): False,
        (0, None): False,
        (1, 0): False,
        (1, None): True,
        (None, 0): False,
        (None, 1): True,
    },
    ("ryy", "c3sx"): False,
    ("ryy", "ccx"): False,
    ("ryy", "cswap"): {
        (0, 1): False,
        (0, 2): False,
        (0, None): False,
        (1, 0): False,
        (1, 2): True,
        (1, None): False,
        (2, 0): False,
        (2, 1): True,
        (2, None): False,
        (None, 0): False,
        (None, 1): False,
        (None, 2): False,
    },
    ("ryy", "ccz"): False,
    ("ryy", "rccx"): False,
    ("ryy", "rcccx"): False,
    ("ryy", "csdg"): False,
    ("ryy", "swap"): {
        (0, 1): True,
        (0, None): False,
        (1, 0): True,
        (1, None): False,
        (None, 0): False,
        (None, 1): False,
    },
    ("ryy", "iswap"): {
        (0, 1): True,
        (0, None): False,
        (1, 0): True,
        (1, None): False,
        (None, 0): False,
        (None, 1): False,
    },
    ("ryy", "ryy"): True,
    ("ryy", "rzz"): {
        (0, 1): True,
        (0, None): False,
        (1, 0): True,
        (1, None): False,
        (None, 0): False,
        (None, 1): False,
    },
    ("ryy", "rzx"): {
        (0, 1): True,
        (0, None): False,
        (1, 0): True,
        (1, None): False,
        (None, 0): False,
        (None, 1): False,
    },
    ("rzz", "c3sx"): {
        (0, 1): True,
        (0, 2): True,
        (0, 3): False,
        (0, None): True,
        (1, 0): True,
        (1, 2): True,
        (1, 3): False,
        (1, None): True,
        (2, 0): True,
        (2, 1): True,
        (2, 3): False,
        (2, None): True,
        (3, 0): False,
        (3, 1): False,
        (3, 2): False,
        (3, None): False,
        (None, 0): True,
        (None, 1): True,
        (None, 2): True,
        (None, 3): False,
    },
    ("rzz", "ccx"): {
        (0, 1): True,
        (0, 2): False,
        (0, None): True,
        (1, 0): True,
        (1, 2): False,
        (1, None): True,
        (2, 0): False,
        (2, 1): False,
        (2, None): False,
        (None, 0): True,
        (None, 1): True,
        (None, 2): False,
    },
    ("rzz", "cswap"): {
        (0, 1): False,
        (0, 2): False,
        (0, None): True,
        (1, 0): False,
        (1, 2): True,
        (1, None): False,
        (2, 0): False,
        (2, 1): True,
        (2, None): False,
        (None, 0): True,
        (None, 1): False,
        (None, 2): False,
    },
    ("rzz", "ccz"): True,
    ("rzz", "rccx"): {
        (0, 1): True,
        (0, 2): False,
        (0, None): True,
        (1, 0): True,
        (1, 2): False,
        (1, None): True,
        (2, 0): False,
        (2, 1): False,
        (2, None): False,
        (None, 0): True,
        (None, 1): True,
        (None, 2): False,
    },
    ("rzz", "rcccx"): {
        (0, 1): True,
        (0, 2): True,
        (0, 3): False,
        (0, None): True,
        (1, 0): True,
        (1, 2): True,
        (1, 3): False,
        (1, None): True,
        (2, 0): True,
        (2, 1): True,
        (2, 3): False,
        (2, None): True,
        (3, 0): False,
        (3, 1): False,
        (3, 2): False,
        (3, None): False,
        (None, 0): True,
        (None, 1): True,
        (None, 2): True,
        (None, 3): False,
    },
    ("rzz", "csdg"): True,
    ("rzz", "swap"): {
        (0, 1): True,
        (0, None): False,
        (1, 0): True,
        (1, None): False,
        (None, 0): False,
        (None, 1): False,
    },
    ("rzz", "iswap"): {
        (0, 1): True,
        (0, None): False,
        (1, 0): True,
        (1, None): False,
        (None, 0): False,
        (None, 1): False,
    },
    ("rzz", "rzz"): True,
    ("rzx", "c3sx"): {
        (0, 1): False,
        (0, 2): False,
        (0, 3): True,
        (0, None): True,
        (1, 0): False,
        (1, 2): False,
        (1, 3): True,
        (1, None): True,
        (2, 0): False,
        (2, 1): False,
        (2, 3): True,
        (2, None): True,
        (3, 0): False,
        (3, 1): False,
        (3, 2): False,
        (3, None): False,
        (None, 0): False,
        (None, 1): False,
        (None, 2): False,
        (None, 3): True,
    },
    ("rzx", "ccx"): {
        (0, 1): False,
        (0, 2): True,
        (0, None): True,
        (1, 0): False,
        (1, 2): True,
        (1, None): True,
        (2, 0): False,
        (2, 1): False,
        (2, None): False,
        (None, 0): False,
        (None, 1): False,
        (None, 2): True,
    },
    ("rzx", "cswap"): {
        (0, 1): False,
        (0, 2): False,
        (0, None): True,
        (1, 0): False,
        (1, 2): False,
        (1, None): False,
        (2, 0): False,
        (2, 1): False,
        (2, None): False,
        (None, 0): False,
        (None, 1): False,
        (None, 2): False,
    },
    ("rzx", "ccz"): {
        (0, 1): False,
        (0, 2): False,
        (0, None): True,
        (1, 0): False,
        (1, 2): False,
        (1, None): True,
        (2, 0): False,
        (2, 1): False,
        (2, None): True,
        (None, 0): False,
        (None, 1): False,
        (None, 2): False,
    },
    ("rzx", "rccx"): {
        (0, 1): False,
        (0, 2): False,
        (0, None): True,
        (1, 0): False,
        (1, 2): False,
        (1, None): True,
        (2, 0): False,
        (2, 1): False,
        (2, None): False,
        (None, 0): False,
        (None, 1): False,
        (None, 2): False,
    },
    ("rzx", "rcccx"): {
        (0, 1): False,
        (0, 2): False,
        (0, 3): False,
        (0, None): True,
        (1, 0): False,
        (1, 2): False,
        (1, 3): False,
        (1, None): True,
        (2, 0): False,
        (2, 1): False,
        (2, 3): False,
        (2, None): True,
        (3, 0): False,
        (3, 1): False,
        (3, 2): False,
        (3, None): False,
        (None, 0): False,
        (None, 1): False,
        (None, 2): False,
        (None, 3): False,
    },
    ("rzx", "csdg"): {
        (0, 1): False,
        (0, None): True,
        (1, 0): False,
        (1, None): True,
        (None, 0): False,
        (None, 1): False,
    },
    ("rzx", "swap"): False,
    ("rzx", "iswap"): False,
    ("rzx", "rzz"): {
        (0, 1): False,
        (0, None): True,
        (1, 0): False,
        (1, None): True,
        (None, 0): False,
        (None, 1): False,
    },
    ("rzx", "rzx"): {
        (0, 1): True,
        (0, None): True,
        (1, 0): True,
        (1, None): False,
        (None, 0): False,
        (None, 1): True,
    },
}



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Tests for qiskit/utils"""

from unittest import mock

from qiskit.utils import multiprocessing
from test import QiskitTestCase  # pylint: disable=wrong-import-order


class TestUtil(QiskitTestCase):
    """Tests for qiskit/_util.py"""

    def test_local_hardware_five_cpu_count(self):
        """Test cpu count is half when sched affinity is 5"""
        with mock.patch.object(multiprocessing, "os"):
            multiprocessing.os.sched_getaffinity = mock.MagicMock(return_value=set(range(5)))
            result = multiprocessing.local_hardware_info()
        self.assertEqual(2, result["cpus"])

    def test_local_hardware_sixty_four_cpu_count(self):
        """Test cpu count is 32 when sched affinity is 64"""
        with mock.patch.object(multiprocessing, "os"):
            multiprocessing.os.sched_getaffinity = mock.MagicMock(return_value=set(range(64)))
            result = multiprocessing.local_hardware_info()
        self.assertEqual(32, result["cpus"])

    def test_local_hardware_no_cpu_count(self):
        """Test cpu count fallback to 1 when true value can't be determined"""
        with mock.patch.object(multiprocessing, "os"):
            multiprocessing.os.sched_getaffinity = mock.MagicMock(return_value=set())
            result = multiprocessing.local_hardware_info()
        self.assertEqual(1, result["cpus"])

    def test_local_hardware_no_sched_five_count(self):
        """Test cpu could if sched affinity method is missing and cpu count is 5."""
        with mock.patch.object(multiprocessing, "os", spec=[]):
            multiprocessing.os.cpu_count = mock.MagicMock(return_value=5)
            del multiprocessing.os.sched_getaffinity
            result = multiprocessing.local_hardware_info()
        self.assertEqual(2, result["cpus"])

    def test_local_hardware_no_sched_sixty_four_count(self):
        """Test cpu could if sched affinity method is missing and cpu count is 64."""
        with mock.patch.object(multiprocessing, "os", spec=[]):
            multiprocessing.os.cpu_count = mock.MagicMock(return_value=64)
            del multiprocessing.os.sched_getaffinity
            result = multiprocessing.local_hardware_info()
        self.assertEqual(32, result["cpus"])

    def test_local_hardware_no_sched_no_count(self):
        """Test cpu count fallback to 1 when no sched getaffinity available."""
        with mock.patch.object(multiprocessing, "os", spec=[]):
            multiprocessing.os.cpu_count = mock.MagicMock(return_value=None)
            del multiprocessing.os.sched_getaffinity
            result = multiprocessing.local_hardware_info()
        self.assertEqual(1, result["cpus"])



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2019.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Superoperator representation of a Quantum Channel."""

from __future__ import annotations

import copy as _copy
import math
from typing import TYPE_CHECKING

import numpy as np

from qiskit import _numpy_compat
from qiskit.circuit.instruction import Instruction
from qiskit.circuit.quantumcircuit import QuantumCircuit
from qiskit.exceptions import QiskitError
from qiskit.quantum_info.operators.base_operator import BaseOperator
from qiskit.quantum_info.operators.channel.quantum_channel import QuantumChannel
from qiskit.quantum_info.operators.channel.transformations import _bipartite_tensor, _to_superop
from qiskit.quantum_info.operators.mixins import generate_apidocs
from qiskit.quantum_info.operators.op_shape import OpShape
from qiskit.quantum_info.operators.operator import Operator

if TYPE_CHECKING:
    from qiskit import circuit
    from qiskit.quantum_info.states.densitymatrix import DensityMatrix
    from qiskit.quantum_info.states.statevector import Statevector


class SuperOp(QuantumChannel):
    r"""Superoperator representation of a quantum channel.

    The Superoperator representation of a quantum channel :math:`\mathcal{E}`
    is a matrix :math:`S` such that the evolution of a
    :class:`~qiskit.quantum_info.DensityMatrix` :math:`\rho` is given by

    .. math::

        |\mathcal{E}(\rho)\rangle\!\rangle = S |\rho\rangle\!\rangle

    where the double-ket notation :math:`|A\rangle\!\rangle` denotes a vector
    formed by stacking the columns of the matrix :math:`A`
    *(column-vectorization)*.

    See reference [1] for further details.

    References:
        1. C.J. Wood, J.D. Biamonte, D.G. Cory, *Tensor networks and graphical calculus
           for open quantum systems*, Quant. Inf. Comp. 15, 0579-0811 (2015).
           `arXiv:1111.6950 [quant-ph] <https://arxiv.org/abs/1111.6950>`_
    """

    def __init__(
        self,
        data: QuantumCircuit | circuit.instruction.Instruction | BaseOperator | np.ndarray,
        input_dims: tuple | None = None,
        output_dims: tuple | None = None,
    ):
        """Initialize a quantum channel Superoperator operator.

        Args:
            data: data to initialize superoperator.
            input_dims: the input subsystem dimensions.
            output_dims: the output subsystem dimensions.

        Raises:
            QiskitError: if input data cannot be initialized as a
                         superoperator.

        Additional Information:
            If the input or output dimensions are None, they will be
            automatically determined from the input data. If the input data is
            a Numpy array of shape (4**N, 4**N) qubit systems will be used. If
            the input operator is not an N-qubit operator, it will assign a
            single subsystem with dimension specified by the shape of the input.
        """
        # If the input is a raw list or matrix we assume that it is
        # already a superoperator.
        if isinstance(data, (list, np.ndarray)):
            # We initialize directly from superoperator matrix
            super_mat = np.asarray(data, dtype=complex)
            # Determine total input and output dimensions
            dout, din = super_mat.shape
            input_dim = int(math.sqrt(din))
            output_dim = int(math.sqrt(dout))
            if output_dim**2 != dout or input_dim**2 != din:
                raise QiskitError("Invalid shape for SuperOp matrix.")
            op_shape = OpShape.auto(
                dims_l=output_dims, dims_r=input_dims, shape=(output_dim, input_dim)
            )
        else:
            # Otherwise we initialize by conversion from another Qiskit
            # object into the QuantumChannel.
            if isinstance(data, (QuantumCircuit, Instruction)):
                # If the input is a Terra QuantumCircuit or Instruction we
                # perform a simulation to construct the circuit superoperator.
                # This will only work if the circuit or instruction can be
                # defined in terms of instructions which have no classical
                # register components. The instructions can be gates, reset,
                # or Kraus instructions. Any conditional gates or measure
                # will cause an exception to be raised.
                data = self._init_instruction(data)
            else:
                # We use the QuantumChannel init transform to initialize
                # other objects into a QuantumChannel or Operator object.
                data = self._init_transformer(data)
            # Now that the input is an operator we convert it to a
            # SuperOp object
            op_shape = data._op_shape
            input_dim, output_dim = data.dim
            rep = getattr(data, "_channel_rep", "Operator")
            super_mat = _to_superop(rep, data._data, input_dim, output_dim)
        # Initialize QuantumChannel
        super().__init__(super_mat, op_shape=op_shape)

    def __array__(self, dtype=None, copy=_numpy_compat.COPY_ONLY_IF_NEEDED):
        dtype = self.data.dtype if dtype is None else dtype
        return np.array(self.data, dtype=dtype, copy=copy)

    @property
    def _tensor_shape(self):
        """Return the tensor shape of the superoperator matrix"""
        return 2 * tuple(reversed(self._op_shape.dims_l())) + 2 * tuple(
            reversed(self._op_shape.dims_r())
        )

    @property
    def _bipartite_shape(self):
        """Return the shape for bipartite matrix"""
        return (self._output_dim, self._output_dim, self._input_dim, self._input_dim)

    # ---------------------------------------------------------------------
    # BaseOperator methods
    # ---------------------------------------------------------------------

    def conjugate(self):
        ret = _copy.copy(self)
        ret._data = np.conj(self._data)
        return ret

    def transpose(self):
        ret = _copy.copy(self)
        ret._data = np.transpose(self._data)
        ret._op_shape = self._op_shape.transpose()
        return ret

    def adjoint(self):
        ret = _copy.copy(self)
        ret._data = np.conj(np.transpose(self._data))
        ret._op_shape = self._op_shape.transpose()
        return ret

    def tensor(self, other: SuperOp) -> SuperOp:
        if not isinstance(other, SuperOp):
            other = SuperOp(other)
        return self._tensor(self, other)

    def expand(self, other: SuperOp) -> SuperOp:
        if not isinstance(other, SuperOp):
            other = SuperOp(other)
        return self._tensor(other, self)

    @classmethod
    def _tensor(cls, a, b):
        ret = _copy.copy(a)
        ret._op_shape = a._op_shape.tensor(b._op_shape)
        ret._data = _bipartite_tensor(
            a._data, b.data, shape1=a._bipartite_shape, shape2=b._bipartite_shape
        )
        return ret

    def compose(self, other: SuperOp, qargs: list | None = None, front: bool = False) -> SuperOp:
        if qargs is None:
            qargs = getattr(other, "qargs", None)
        if not isinstance(other, SuperOp):
            other = SuperOp(other)
        # Validate dimensions are compatible and return the composed
        # operator dimensions
        new_shape = self._op_shape.compose(other._op_shape, qargs, front)
        input_dims = new_shape.dims_r()
        output_dims = new_shape.dims_l()

        # Full composition of superoperators
        if qargs is None:
            if front:
                data = np.dot(self._data, other.data)
            else:
                data = np.dot(other.data, self._data)
            ret = SuperOp(data, input_dims, output_dims)
            ret._op_shape = new_shape
            return ret

        # Compute tensor contraction indices from qargs
        num_qargs_l, num_qargs_r = self._op_shape.num_qargs
        if front:
            num_indices = num_qargs_r
            shift = 2 * num_qargs_l
            right_mul = True
        else:
            num_indices = num_qargs_l
            shift = 0
            right_mul = False

        # Reshape current matrix
        # Note that we must reverse the subsystem dimension order as
        # qubit 0 corresponds to the right-most position in the tensor
        # product, which is the last tensor wire index.
        tensor = np.reshape(self.data, self._tensor_shape)
        mat = np.reshape(other.data, other._tensor_shape)
        # Add first set of indices
        indices = [2 * num_indices - 1 - qubit for qubit in qargs] + [
            num_indices - 1 - qubit for qubit in qargs
        ]
        final_shape = [np.prod(output_dims) ** 2, np.prod(input_dims) ** 2]
        data = np.reshape(
            Operator._einsum_matmul(tensor, mat, indices, shift, right_mul), final_shape
        )
        ret = SuperOp(data, input_dims, output_dims)
        ret._op_shape = new_shape
        return ret

    # ---------------------------------------------------------------------
    # Additional methods
    # ---------------------------------------------------------------------

    def _evolve(self, state, qargs=None):
        """Evolve a quantum state by the quantum channel.

        Args:
            state (DensityMatrix or Statevector): The input state.
            qargs (list): a list of quantum state subsystem positions to apply
                           the quantum channel on.

        Returns:
            DensityMatrix: the output quantum state as a density matrix.

        Raises:
            QiskitError: if the quantum channel dimension does not match the
                         specified quantum state subsystem dimensions.
        """
        # Prevent cyclic imports by importing DensityMatrix here
        # pylint: disable=cyclic-import
        from qiskit.quantum_info.states.densitymatrix import DensityMatrix

        if not isinstance(state, DensityMatrix):
            state = DensityMatrix(state)

        if qargs is None:
            # Evolution on full matrix
            if state._op_shape.shape[0] != self._op_shape.shape[1]:
                raise QiskitError(
                    "Operator input dimension is not equal to density matrix dimension."
                )
            # We reshape in column-major vectorization (Fortran order in Numpy)
            # since that is how the SuperOp is defined
            vec = np.ravel(state.data, order="F")
            mat = np.reshape(
                np.dot(self.data, vec), (self._output_dim, self._output_dim), order="F"
            )
            return DensityMatrix(mat, dims=self.output_dims())
        # Otherwise we are applying an operator only to subsystems
        # Check dimensions of subsystems match the operator
        if state.dims(qargs) != self.input_dims():
            raise QiskitError(
                "Operator input dimensions are not equal to statevector subsystem dimensions."
            )
        # Reshape statevector and operator
        tensor = np.reshape(state.data, state._op_shape.tensor_shape)
        mat = np.reshape(self.data, self._tensor_shape)
        # Construct list of tensor indices of statevector to be contracted
        num_indices = len(state.dims())
        indices = [num_indices - 1 - qubit for qubit in qargs] + [
            2 * num_indices - 1 - qubit for qubit in qargs
        ]
        tensor = Operator._einsum_matmul(tensor, mat, indices)
        # Replace evolved dimensions
        new_dims = list(state.dims())
        output_dims = self.output_dims()
        for i, qubit in enumerate(qargs):
            new_dims[qubit] = output_dims[i]
        new_dim = np.prod(new_dims)
        # reshape tensor to density matrix
        tensor = np.reshape(tensor, (new_dim, new_dim))
        return DensityMatrix(tensor, dims=new_dims)

    @classmethod
    def _init_instruction(cls, instruction):
        """Convert a QuantumCircuit or Instruction to a SuperOp."""
        # Convert circuit to an instruction
        if isinstance(instruction, QuantumCircuit):
            instruction = instruction.to_instruction()
        # Initialize an identity superoperator of the correct size
        # of the circuit
        op = SuperOp(np.eye(4**instruction.num_qubits))
        op._append_instruction(instruction)
        return op

    @classmethod
    def _instruction_to_superop(cls, obj):
        """Return superop for instruction if defined or None otherwise."""
        if not isinstance(obj, Instruction):
            raise QiskitError("Input is not an instruction.")
        chan = None
        if obj.name == "reset":
            # For superoperator evolution we can simulate a reset as
            # a non-unitary superoperator matrix
            chan = SuperOp(np.array([[1, 0, 0, 1], [0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0]]))
        if obj.name == "kraus":
            kraus = obj.params
            dim = len(kraus[0])
            chan = SuperOp(_to_superop("Kraus", (kraus, None), dim, dim))
        elif hasattr(obj, "to_matrix"):
            # If instruction is a gate first we see if it has a
            # `to_matrix` definition and if so use that.
            try:
                kraus = [obj.to_matrix()]
                dim = len(kraus[0])
                chan = SuperOp(_to_superop("Kraus", (kraus, None), dim, dim))
            except QiskitError:
                pass
        return chan

    def _append_instruction(self, obj, qargs=None):
        """Update the current Operator by apply an instruction."""
        from qiskit.circuit.barrier import Barrier

        chan = self._instruction_to_superop(obj)
        if chan is not None:
            # Perform the composition and inplace update the current state
            # of the operator
            op = self.compose(chan, qargs=qargs)
            self._data = op.data
        elif isinstance(obj, Barrier):
            return
        else:
            # If the instruction doesn't have a matrix defined we use its
            # circuit decomposition definition if it exists, otherwise we
            # cannot compose this gate and raise an error.
            if obj.definition is None:
                raise QiskitError(f"Cannot apply Instruction: {obj.name}")
            if not isinstance(obj.definition, QuantumCircuit):
                raise QiskitError(
                    f"{obj.name} instruction definition is {type(obj.definition)}; "
                    "expected QuantumCircuit"
                )
            qubit_indices = {bit: idx for idx, bit in enumerate(obj.definition.qubits)}
            for instruction in obj.definition.data:
                if instruction.clbits:
                    raise QiskitError(
                        "Cannot apply instruction with classical bits:"
                        f" {instruction.operation.name}"
                    )
                # Get the integer position of the flat register
                if qargs is None:
                    new_qargs = [qubit_indices[tup] for tup in instruction.qubits]
                else:
                    new_qargs = [qargs[qubit_indices[tup]] for tup in instruction.qubits]
                self._append_instruction(instruction.operation, qargs=new_qargs)


# Update docstrings for API docs
generate_apidocs(SuperOp)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2020
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Custom Lazy Iterator class
"""

from __future__ import annotations
from abc import ABC, abstractmethod


class CustomIterator(ABC):
    """Lazy custom iteration and item access."""

    def __init__(self, obj):
        self.obj = obj
        self._iter = 0

    @abstractmethod
    def __getitem__(self, key):
        """Get next item"""
        # This method should be overridden for lazy conversion of
        # iterator only at a given key value
        pass

    def __repr__(self):
        return f"<{type(self.obj)}_iterator at {hex(id(self))}>"

    def __len__(self):
        return len(self.obj)

    def __iter__(self):
        self._iter = 0
        return self

    def __next__(self):
        if self._iter >= len(self):
            raise StopIteration
        self._iter += 1
        return self[self._iter - 1]



# This code is part of Qiskit.
#
# (C) Copyright IBM 2023
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable=no-member,invalid-name,missing-docstring,no-name-in-module
# pylint: disable=attribute-defined-outside-init,unsubscriptable-object

import os

from qiskit.compiler import transpile
from qiskit import QuantumCircuit
from qiskit.transpiler import InstructionDurations
from qiskit.providers.fake_provider import Fake20QV1

from .utils import build_qv_model_circuit


class TranspilerLevelBenchmarks:
    params = [0, 1, 2, 3]
    param_names = ["transpiler optimization level"]
    timeout = 600

    def setup(self, _):
        self.rochester_coupling_map = [
            [0, 5],
            [0, 1],
            [1, 2],
            [1, 0],
            [2, 3],
            [2, 1],
            [3, 4],
            [3, 2],
            [4, 6],
            [4, 3],
            [5, 9],
            [5, 0],
            [6, 13],
            [6, 4],
            [7, 16],
            [7, 8],
            [8, 9],
            [8, 7],
            [9, 10],
            [9, 8],
            [9, 5],
            [10, 11],
            [10, 9],
            [11, 17],
            [11, 12],
            [11, 10],
            [12, 13],
            [12, 11],
            [13, 14],
            [13, 12],
            [13, 6],
            [14, 15],
            [14, 13],
            [15, 18],
            [15, 14],
            [16, 19],
            [16, 7],
            [17, 23],
            [17, 11],
            [18, 27],
            [18, 15],
            [19, 20],
            [19, 16],
            [20, 21],
            [20, 19],
            [21, 28],
            [21, 22],
            [21, 20],
            [22, 23],
            [22, 21],
            [23, 24],
            [23, 22],
            [23, 17],
            [24, 25],
            [24, 23],
            [25, 29],
            [25, 26],
            [25, 24],
            [26, 27],
            [26, 25],
            [27, 26],
            [27, 18],
            [28, 32],
            [28, 21],
            [29, 36],
            [29, 25],
            [30, 39],
            [30, 31],
            [31, 32],
            [31, 30],
            [32, 33],
            [32, 31],
            [32, 28],
            [33, 34],
            [33, 32],
            [34, 40],
            [34, 35],
            [34, 33],
            [35, 36],
            [35, 34],
            [36, 37],
            [36, 35],
            [36, 29],
            [37, 38],
            [37, 36],
            [38, 41],
            [38, 37],
            [39, 42],
            [39, 30],
            [40, 46],
            [40, 34],
            [41, 50],
            [41, 38],
            [42, 43],
            [42, 39],
            [43, 44],
            [43, 42],
            [44, 51],
            [44, 45],
            [44, 43],
            [45, 46],
            [45, 44],
            [46, 47],
            [46, 45],
            [46, 40],
            [47, 48],
            [47, 46],
            [48, 52],
            [48, 49],
            [48, 47],
            [49, 50],
            [49, 48],
            [50, 49],
            [50, 41],
            [51, 44],
            [52, 48],
        ]
        self.basis_gates = ["u1", "u2", "u3", "cx", "id"]
        self.qv_50_x_20 = build_qv_model_circuit(50, 20, 0)
        self.qv_14_x_14 = build_qv_model_circuit(14, 14, 0)
        self.qasm_path = os.path.abspath(os.path.join(os.path.dirname(__file__), "qasm"))
        large_qasm_path = os.path.join(self.qasm_path, "test_eoh_qasm.qasm")
        self.large_qasm = QuantumCircuit.from_qasm_file(large_qasm_path)
        self.melbourne = Fake20QV1()
        self.durations = InstructionDurations(
            [
                ("u1", None, 0),
                ("id", None, 160),
                ("u2", None, 160),
                ("u3", None, 320),
                ("cx", None, 800),
                ("measure", None, 3200),
            ],
            dt=1e-9,
        )

    def time_quantum_volume_transpile_50_x_20(self, transpiler_level):
        transpile(
            self.qv_50_x_20,
            basis_gates=self.basis_gates,
            coupling_map=self.rochester_coupling_map,
            seed_transpiler=0,
            optimization_level=transpiler_level,
        )

    def track_depth_quantum_volume_transpile_50_x_20(self, transpiler_level):
        return transpile(
            self.qv_50_x_20,
            basis_gates=self.basis_gates,
            coupling_map=self.rochester_coupling_map,
            seed_transpiler=0,
            optimization_level=transpiler_level,
        ).depth()

    def time_transpile_from_large_qasm(self, transpiler_level):
        transpile(
            self.large_qasm,
            basis_gates=self.basis_gates,
            coupling_map=self.rochester_coupling_map,
            seed_transpiler=0,
            optimization_level=transpiler_level,
        )

    def track_depth_transpile_from_large_qasm(self, transpiler_level):
        return transpile(
            self.large_qasm,
            basis_gates=self.basis_gates,
            coupling_map=self.rochester_coupling_map,
            seed_transpiler=0,
            optimization_level=transpiler_level,
        ).depth()

    def time_transpile_from_large_qasm_backend_with_prop(self, transpiler_level):
        transpile(
            self.large_qasm, self.melbourne, seed_transpiler=0, optimization_level=transpiler_level
        )

    def track_depth_transpile_from_large_qasm_backend_with_prop(self, transpiler_level):
        return transpile(
            self.large_qasm, self.melbourne, seed_transpiler=0, optimization_level=transpiler_level
        ).depth()

    def time_transpile_qv_14_x_14(self, transpiler_level):
        transpile(
            self.qv_14_x_14, self.melbourne, seed_transpiler=0, optimization_level=transpiler_level
        )

    def track_depth_transpile_qv_14_x_14(self, transpiler_level):
        return transpile(
            self.qv_14_x_14, self.melbourne, seed_transpiler=0, optimization_level=transpiler_level
        ).depth()

    def time_schedule_qv_14_x_14(self, transpiler_level):
        transpile(
            self.qv_14_x_14,
            self.melbourne,
            seed_transpiler=0,
            optimization_level=transpiler_level,
            scheduling_method="alap",
            instruction_durations=self.durations,
        )

    # limit optimization levels to reduce time
    time_schedule_qv_14_x_14.params = [0, 1]



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.


"""Various ways to divide a DAG into blocks of nodes, to split blocks of nodes
into smaller sub-blocks, and to consolidate blocks."""
from __future__ import annotations

from collections.abc import Iterable, Callable

from qiskit.dagcircuit import DAGDepNode

from qiskit.circuit import QuantumCircuit, CircuitInstruction, ClassicalRegister, Bit
from qiskit.circuit.controlflow import condition_resources
from . import DAGOpNode, DAGCircuit, DAGDependency
from .exceptions import DAGCircuitError


class BlockCollector:
    """This class implements various strategies of dividing a DAG (direct acyclic graph)
    into blocks of nodes that satisfy certain criteria. It works both with the
    :class:`~qiskit.dagcircuit.DAGCircuit` and
    :class:`~qiskit.dagcircuit.DAGDependency` representations of a DAG, where
    DagDependency takes into account commutativity between nodes.

    Collecting nodes from DAGDependency generally leads to more optimal results, but is
    slower, as it requires to construct a DAGDependency beforehand. Thus, DAGCircuit should
    be used with lower transpiler settings, and DAGDependency should be used with higher
    transpiler settings.

    In general, there are multiple ways to collect maximal blocks. The approaches used
    here are of the form 'starting from the input nodes of a DAG, greedily collect
    the largest block of nodes that match certain criteria'. For additional details,
    see https://github.com/Qiskit/qiskit-terra/issues/5775.
    """

    def __init__(self, dag: DAGCircuit | DAGDependency):
        """
        Args:
            dag (Union[DAGCircuit, DAGDependency]): The input DAG.

        Raises:
            DAGCircuitError: the input object is not a DAG.
        """

        self.dag = dag
        self._pending_nodes: list[DAGOpNode | DAGDepNode] | None = None
        self._in_degree: dict[DAGOpNode | DAGDepNode, int] | None = None
        self._collect_from_back = False

        if isinstance(dag, DAGCircuit):
            self.is_dag_dependency = False

        elif isinstance(dag, DAGDependency):
            self.is_dag_dependency = True

        else:
            raise DAGCircuitError("not a DAG.")

    def _setup_in_degrees(self):
        """For an efficient implementation, for every node we keep the number of its
        unprocessed immediate predecessors (called ``_in_degree``). This ``_in_degree``
        is set up at the start and updated throughout the algorithm.
        A node is leaf (or input) node iff its ``_in_degree`` is 0.
        When a node is (marked as) collected, the ``_in_degree`` of each of its immediate
        successor is updated by subtracting 1.
        Additionally, ``_pending_nodes`` explicitly keeps the list of nodes whose
        ``_in_degree`` is 0.
        """
        self._pending_nodes = []
        self._in_degree = {}
        for node in self._op_nodes():
            deg = len(self._direct_preds(node))
            self._in_degree[node] = deg
            if deg == 0:
                self._pending_nodes.append(node)

    def _op_nodes(self) -> Iterable[DAGOpNode | DAGDepNode]:
        """Returns DAG nodes."""
        if not self.is_dag_dependency:
            return self.dag.op_nodes()
        else:
            return self.dag.get_nodes()

    def _direct_preds(self, node):
        """Returns direct predecessors of a node. This function takes into account the
        direction of collecting blocks, that is node's predecessors when collecting
        backwards are the direct successors of a node in the DAG.
        """
        if not self.is_dag_dependency:
            if self._collect_from_back:
                return [pred for pred in self.dag.successors(node) if isinstance(pred, DAGOpNode)]
            else:
                return [pred for pred in self.dag.predecessors(node) if isinstance(pred, DAGOpNode)]
        else:
            if self._collect_from_back:
                return [
                    self.dag.get_node(pred_id)
                    for pred_id in self.dag.direct_successors(node.node_id)
                ]
            else:
                return [
                    self.dag.get_node(pred_id)
                    for pred_id in self.dag.direct_predecessors(node.node_id)
                ]

    def _direct_succs(self, node):
        """Returns direct successors of a node. This function takes into account the
        direction of collecting blocks, that is node's successors when collecting
        backwards are the direct predecessors of a node in the DAG.
        """
        if not self.is_dag_dependency:
            if self._collect_from_back:
                return [succ for succ in self.dag.predecessors(node) if isinstance(succ, DAGOpNode)]
            else:
                return [succ for succ in self.dag.successors(node) if isinstance(succ, DAGOpNode)]
        else:
            if self._collect_from_back:
                return [
                    self.dag.get_node(succ_id)
                    for succ_id in self.dag.direct_predecessors(node.node_id)
                ]
            else:
                return [
                    self.dag.get_node(succ_id)
                    for succ_id in self.dag.direct_successors(node.node_id)
                ]

    def _have_uncollected_nodes(self):
        """Returns whether there are uncollected (pending) nodes"""
        return len(self._pending_nodes) > 0

    def collect_matching_block(self, filter_fn: Callable) -> list[DAGOpNode | DAGDepNode]:
        """Iteratively collects the largest block of input nodes (that is, nodes with
        ``_in_degree`` equal to 0) that match a given filtering function.
        Examples of this include collecting blocks of swap gates,
        blocks of linear gates (CXs and SWAPs), blocks of Clifford gates, blocks of single-qubit gates,
        blocks of two-qubit gates, etc.  Here 'iteratively' means that once a node is collected,
        the ``_in_degree`` of each of its immediate successor is decreased by 1, allowing more nodes
        to become input and to be eligible for collecting into the current block.
        Returns the block of collected nodes.
        """
        current_block = []
        unprocessed_pending_nodes = self._pending_nodes
        self._pending_nodes = []

        # Iteratively process unprocessed_pending_nodes:
        # - any node that does not match filter_fn is added to pending_nodes
        # - any node that match filter_fn is added to the current_block,
        #   and some of its successors may be moved to unprocessed_pending_nodes.
        while unprocessed_pending_nodes:
            new_pending_nodes = []
            for node in unprocessed_pending_nodes:
                if filter_fn(node):
                    current_block.append(node)

                    # update the _in_degree of node's successors
                    for suc in self._direct_succs(node):
                        self._in_degree[suc] -= 1
                        if self._in_degree[suc] == 0:
                            new_pending_nodes.append(suc)
                else:
                    self._pending_nodes.append(node)
            unprocessed_pending_nodes = new_pending_nodes

        return current_block

    def collect_all_matching_blocks(
        self,
        filter_fn,
        split_blocks=True,
        min_block_size=2,
        split_layers=False,
        collect_from_back=False,
    ):
        """Collects all blocks that match a given filtering function filter_fn.
        This iteratively finds the largest block that does not match filter_fn,
        then the largest block that matches filter_fn, and so on, until no more uncollected
        nodes remain. Intuitively, finding larger blocks of non-matching nodes helps to
        find larger blocks of matching nodes later on.

        After the blocks are collected, they can be optionally refined. The option
        ``split_blocks`` allows to split collected blocks into sub-blocks over disjoint
        qubit subsets. The option ``split_layers`` allows to split collected blocks
        into layers of non-overlapping instructions. The option ``min_block_size``
        specifies the minimum number of gates in the block for the block to be collected.

        By default, blocks are collected in the direction from the inputs towards the outputs
        of the circuit. The option ``collect_from_back`` allows to change this direction,
        that is collect blocks from the outputs towards the inputs of the circuit.

        Returns the list of matching blocks only.
        """

        def not_filter_fn(node):
            """Returns the opposite of filter_fn."""
            return not filter_fn(node)

        # Note: the collection direction must be specified before setting in-degrees
        self._collect_from_back = collect_from_back
        self._setup_in_degrees()

        # Iteratively collect non-matching and matching blocks.
        matching_blocks: list[list[DAGOpNode | DAGDepNode]] = []
        while self._have_uncollected_nodes():
            self.collect_matching_block(not_filter_fn)
            matching_block = self.collect_matching_block(filter_fn)
            if matching_block:
                matching_blocks.append(matching_block)

        # If the option split_layers is set, refine blocks by splitting them into layers
        # of non-overlapping instructions (in other words, into depth-1 sub-blocks).
        if split_layers:
            tmp_blocks = []
            for block in matching_blocks:
                tmp_blocks.extend(split_block_into_layers(block))
            matching_blocks = tmp_blocks

        # If the option split_blocks is set, refine blocks by splitting them into sub-blocks over
        # disconnected qubit subsets.
        if split_blocks:
            tmp_blocks = []
            for block in matching_blocks:
                tmp_blocks.extend(BlockSplitter().run(block))
            matching_blocks = tmp_blocks

        # If we are collecting from the back, both the order of the blocks
        # and the order of nodes in each block should be reversed.
        if self._collect_from_back:
            matching_blocks = [block[::-1] for block in matching_blocks[::-1]]

        # Keep only blocks with at least min_block_sizes.
        matching_blocks = [block for block in matching_blocks if len(block) >= min_block_size]

        return matching_blocks


class BlockSplitter:
    """Splits a block of nodes into sub-blocks over disjoint qubits.
    The implementation is based on the Disjoint Set Union data structure."""

    def __init__(self):
        self.leader = {}  # qubit's group leader
        self.group = {}  # qubit's group

    def find_leader(self, index):
        """Find in DSU."""
        if index not in self.leader:
            self.leader[index] = index
            self.group[index] = []
            return index
        if self.leader[index] == index:
            return index
        self.leader[index] = self.find_leader(self.leader[index])
        return self.leader[index]

    def union_leaders(self, index1, index2):
        """Union in DSU."""
        leader1 = self.find_leader(index1)
        leader2 = self.find_leader(index2)
        if leader1 == leader2:
            return
        if len(self.group[leader1]) < len(self.group[leader2]):
            leader1, leader2 = leader2, leader1

        self.leader[leader2] = leader1
        self.group[leader1].extend(self.group[leader2])
        self.group[leader2].clear()

    def run(self, block):
        """Splits block of nodes into sub-blocks over disjoint qubits."""
        for node in block:
            indices = node.qargs
            if not indices:
                continue
            first = indices[0]
            for index in indices[1:]:
                self.union_leaders(first, index)
            self.group[self.find_leader(first)].append(node)

        blocks = []
        for index, item in self.leader.items():
            if index == item:
                blocks.append(self.group[index])

        return blocks


def split_block_into_layers(block: list[DAGOpNode | DAGDepNode]):
    """Splits a block of nodes into sub-blocks of non-overlapping instructions
    (or, in other words, into depth-1 sub-blocks).
    """
    bit_depths: dict[Bit, int] = {}
    layers: list[list[DAGOpNode | DAGDepNode]] = []

    for node in block:
        cur_bits = set(node.qargs)
        cur_bits.update(node.cargs)

        cond = getattr(node.op, "_condition", None)
        if cond is not None:
            cur_bits.update(condition_resources(cond).clbits)

        cur_depth = max(bit_depths.get(bit, 0) for bit in cur_bits)
        while len(layers) <= cur_depth:
            layers.append([])

        for bit in cur_bits:
            bit_depths[bit] = cur_depth + 1
        layers[cur_depth].append(node)

    return layers


class BlockCollapser:
    """This class implements various strategies of consolidating blocks of nodes
     in a DAG (direct acyclic graph). It works both with the
    :class:`~qiskit.dagcircuit.DAGCircuit` and
    :class:`~qiskit.dagcircuit.DAGDependency` DAG representations.
    """

    def __init__(self, dag):
        """
        Args:
            dag (Union[DAGCircuit, DAGDependency]): The input DAG.
        """

        self.dag = dag

    def collapse_to_operation(self, blocks, collapse_fn):
        """For each block, constructs a quantum circuit containing instructions in the block,
        then uses collapse_fn to collapse this circuit into a single operation.
        """
        global_index_map = {wire: idx for idx, wire in enumerate(self.dag.qubits)}
        global_index_map.update({wire: idx for idx, wire in enumerate(self.dag.clbits)})

        for block in blocks:
            # Find the sets of qubits/clbits used in this block (which might be much smaller
            # than the set of all qubits/clbits).
            cur_qubits = set()
            cur_clbits = set()

            # Additionally, find the set of classical registers used in conditions over full registers
            # (in such a case, we need to add that register to the block circuit, not just its clbits).
            cur_clregs = set()

            for node in block:
                cur_qubits.update(node.qargs)
                cur_clbits.update(node.cargs)
                cond = getattr(node.op, "_condition", None)
                if cond is not None:
                    cur_clbits.update(condition_resources(cond).clbits)
                    if isinstance(cond[0], ClassicalRegister):
                        cur_clregs.add(cond[0])

            # For reproducibility, order these qubits/clbits compatibly with the global order.
            sorted_qubits = sorted(cur_qubits, key=lambda x: global_index_map[x])
            sorted_clbits = sorted(cur_clbits, key=lambda x: global_index_map[x])

            qc = QuantumCircuit(sorted_qubits, sorted_clbits)

            # Add classical registers used in conditions over registers
            for reg in cur_clregs:
                qc.add_register(reg)

            # Construct a quantum circuit from the nodes in the block, remapping the qubits.
            wire_pos_map = {qb: ix for ix, qb in enumerate(sorted_qubits)}
            wire_pos_map.update({qb: ix for ix, qb in enumerate(sorted_clbits)})

            for node in block:
                instructions = qc.append(CircuitInstruction(node.op, node.qargs, node.cargs))
                cond = getattr(node.op, "_condition", None)
                if cond is not None:
                    instructions.c_if(*cond)

            # Collapse this quantum circuit into an operation.
            op = collapse_fn(qc)

            # Replace the block of nodes in the DAG by the constructed operation
            # (the function replace_block_with_op is implemented both in DAGCircuit and DAGDependency).
            self.dag.replace_block_with_op(block, op, wire_pos_map, cycle_check=False)
        return self.dag
